import os
import unittest

from ayeaye import AccessMode
from ayeaye.connect import Connect
from ayeaye.connectors.csv_connector import CsvConnector, TsvConnector
from ayeaye.connectors.fake import FakeDataConnector
from ayeaye.connectors.sqlalchemy_database import SqlAlchemyDatabaseConnector

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_CSV_PATH = os.path.join(PROJECT_TEST_PATH, "data", "deadly_creatures.csv")
EXAMPLE_TSV_PATH = os.path.join(PROJECT_TEST_PATH, "data", "monkeys.tsv")


class AbstractFakeModel:
    def __init__(self):
        self._connections = {}


class TestConnect(unittest.TestCase):
    def test_connect_standalone(self):
        """
        :class:`ayeaye.Connect` can be used outside of the ETL so data discovery can use the same
        way of working as full :class:`ayeaye.Model`s.
        """
        # happy path
        # it works without Connect being part of a ayeaye.Model
        c = Connect(engine_url="fake://MyDataset")
        self.assertEqual({"fake": "data"}, c.data[0])

    def test_connect_spare_kwargs(self):
        """
        subclasses of :class:`ayeaye.connectors.base.DataConnector` can be given specific/custom
        kwargs. An exception should be raised when unclaimed spare kwargs remain. This will make
        it harder for users to make mistakes and typos referring to arguments that never come
        into play.
        """
        c = Connect(engine_url="fake://foo", doesntexist="oh dear")
        with self.assertRaises(ValueError):
            # the kwargs are not used until an engine_url is needed
            c._prepare_connection()

    def test_connect_within_instantiated_class(self):
        """
        Connect used as a class variable. The parent class, which in practice will be a
        :class:`ayeaye.Model`.
        When used as a class variable in an instantiated class, Connect() will store information
        about the dataset within the parent (i.e. Model) class.
        """

        class FakeModel(AbstractFakeModel):
            insects = Connect(engine_url="fake://bugsDB")

        e0 = FakeModel()
        self.assertEqual(0, len(e0._connections))

        # connect on demand/access
        self.assertIsNotNone(e0.insects)
        self.assertEqual(1, len(e0._connections))

    def test_custom_kwargs_are_passed(self):
        """
        ayeaye.Connect should relay kwargs to subclasses of DataConnecter
        """
        # using bigquery because it has custom 'credentials' kwarg
        engine_url = "bigquery://projectId=my_project;datasetId=nice_food;tableId=cakes;"
        c = Connect(engine_url=engine_url, credentials="hello_world")
        # on demand connection
        self.assertIsNotNone(c.data)
        self.assertEqual("hello_world", c._standalone_connection.credentials)

    def test_overlay_args(self):
        """
        Make an access=AccessMode.READ connection in a model into access=AccessMode.WRITE.
        The engine_url stays the same.
        """

        class FakeModel(AbstractFakeModel):
            insects = Connect(engine_url="fake://bugsDB")

        class FakeModelWrite:
            insects = FakeModel.insects.clone(access=AccessMode.WRITE)

            def __init__(self):
                self._connections = {}

        f = FakeModelWrite()
        self.assertEqual("fake://bugsDB", f.insects.engine_url)
        self.assertEqual(AccessMode.WRITE, f.insects.access)

    def test_replace_existing_connect(self):
        class FakeModel(AbstractFakeModel):
            insects = Connect(engine_url="fake://bugsDB")

        m = FakeModel()
        with self.assertRaises(ValueError) as context:
            m.insects = "this is a string, not an instance of Connect"
        self.assertEqual("Only Connect instances can be set", str(context.exception))

        self.assertEqual({}, m._connections, "Connections not initialised prior to access")
        self.assertEqual("fake://bugsDB", m.insects.engine_url, "Original connection")

        m.insects = Connect(engine_url="fake://creepyCrawliesDB")
        self.assertEqual("fake://creepyCrawliesDB", m.insects.engine_url, "New connection")

    def test_update_by_replacement(self):
        """
        Take a connection from a model, make a small tweak and set it back into the model.
        Note is isn't a class tweak (that is tested elsewhere), it's on instances.
        """

        class FakeModel(AbstractFakeModel):
            insects = Connect(engine_url="fake://bugsDB")

        m = FakeModel()
        self.assertTrue(AccessMode.READ == m.insects.access, "Expected starting state not found")

        connect_refs = [k for k in m._connections.keys()]

        # change something, this could have been more dramatic, the engine type for example
        # Assigning this back to insects will re-prepare the connection
        c = m.insects.connect_instance
        c.update(access=AccessMode.WRITE)
        m.insects = c

        self.assertEqual(AccessMode.WRITE, m.insects.access, "Change to connection went missing")
        connect_refs_now = [k for k in m._connections.keys()]
        self.assertEqual(connect_refs, connect_refs_now, "Connect instances shouldn't change")

    def test_update_inline(self):
        """
        A change to the `Connect` propagates to build a new DataConnector with a different type.
        """

        class FakeModel(AbstractFakeModel):
            insects = Connect(engine_url="fake://bugsDB")

        m = FakeModel()
        self.assertIsInstance(m.insects, FakeDataConnector)

        m.insects.update(engine_url="mysql://")
        self.assertIsInstance(m.insects, SqlAlchemyDatabaseConnector)

    def test_compile_time_multiple_engine_urls(self):
        """
        engine_url could be a list of engine_urls.
        In the future, a dictionary version might be added
        """
        tsv_engine_url = "tsv://" + EXAMPLE_TSV_PATH
        csv_engine_url = "csv://" + EXAMPLE_CSV_PATH
        c = Connect(engine_url=[tsv_engine_url, csv_engine_url])

        all_the_animals = []
        for index, data_connector in enumerate(c):
            if index == 0:
                self.assertIsInstance(data_connector, TsvConnector)
            elif index == 1:
                self.assertIsInstance(data_connector, CsvConnector)
            else:
                raise ValueError("Connect has more than expected data connectors")

            all_the_animals += [animal.common_name for animal in data_connector]

        expected = [
            "Goeldi's marmoset",
            "Common squirrel monkey",
            "Crab-eating macaque",
            "Crown of thorns starfish",
            "Golden dart frog",
        ]
        self.assertEqual(expected, all_the_animals)

    def test_compile_time_multiple_engine_urls_in_a_model(self):
        """
        Check the descriptors are treating children of MultiConnector the same as
        the other DataConnector subclasses.
        """

        class AnimalsModel(AbstractFakeModel):
            animals = Connect(engine_url=["tsv://" + EXAMPLE_TSV_PATH, "csv://" + EXAMPLE_CSV_PATH])

        m = AnimalsModel()

        all_the_animals = []
        for animal_dataset in m.animals:
            all_the_animals += [animal.common_name for animal in animal_dataset]

        expected = [
            "Goeldi's marmoset",
            "Common squirrel monkey",
            "Crab-eating macaque",
            "Crown of thorns starfish",
            "Golden dart frog",
        ]
        self.assertEqual(expected, all_the_animals)

    def test_callable_engine_url(self):
        def pointlessly_deterministic_example_callable():
            return "fake://MyDataset"

        c = Connect(engine_url=pointlessly_deterministic_example_callable)
        self.assertEqual({"fake": "data"}, c.data[0], "Example data not found")

    def test_split_instance_and_connect(self):
        """
        Connect() is only called when the class is imported. If the instance of Connect is mutated
        this shouldn't be passed to other instances.
        Bug found in other project.
        """

        class AnimalsModel(AbstractFakeModel):
            animals = Connect(engine_url=[])

        m1 = AnimalsModel()
        m1.animals.engine_url.append("hi")

        m2 = AnimalsModel()
        self.assertEqual([], m2.animals.engine_url)

    def test_steal_model_connect(self):
        """
        Take a Connect from a model's class variables and use it as a standalone Connect.
        """

        class AnimalsModel(AbstractFakeModel):
            animals = Connect(engine_url="csv://" + EXAMPLE_CSV_PATH)

        animals = AnimalsModel.animals
        self.assertEqual("ConnectBind.NEW", str(animals.connection_bind))

        all_the_animals = [animal.common_name for animal in animals]

        expected = ["Crown of thorns starfish", "Golden dart frog"]
        self.assertEqual(expected, all_the_animals)

        self.assertEqual("ConnectBind.STANDALONE", str(animals.connection_bind))

    def test_standalone_in_non_model_instance(self):
        """
        Make changes to a Connect that doesn't belong to an ayeaye.Model.
        Also see :method:`TestModels.test_double_usage` for how this behaves with an ayeaye.Model.
        """

        class NonModelClass:
            "doesn't have self._connections"

            def __init__(self):
                # the instance of Connect is just a variable, it's not an attribute so it's descriptor
                # methods aren't called.
                self.animals = Connect(engine_url="csv://" + EXAMPLE_CSV_PATH)

            def go(self):
                all_the_animals = [animal.common_name for animal in self.animals]
                return all_the_animals

        nmc = NonModelClass()
        self.assertEqual(["Crown of thorns starfish", "Golden dart frog"], nmc.go())

        nmc.animals.update(engine_url="tsv://" + EXAMPLE_TSV_PATH)
        expected_values = ["Goeldi's marmoset", "Common squirrel monkey", "Crab-eating macaque"]
        self.assertEqual(expected_values, nmc.go())

    def test_standalone_as_proxy(self):
        """
        Access an attribute of the subclass that doesn't belong to the DataConnector abstract class.
        """
        animals = Connect(engine_url="csv://" + EXAMPLE_CSV_PATH + ";encoding=magic_encoding")
        self.assertEqual("magic_encoding", animals.encoding)

    def test_construction_args(self):
        with self.assertRaises(ValueError, msg="Ref and engine_url are mutually exclusive"):
            Connect(ref="x", engine_url="tsv://" + EXAMPLE_TSV_PATH)

    def test_connect_callable_kwargs(self):
        """
        :class:`ayeaye.connectots.fake.FakeDataConnector` has an optional kwarg-
        'quantum_accelerator_module' set this using a literal or a callable.
        """

        c = Connect(engine_url="fake://MyDataset", quantum_accelerator_module="entanglement_v1")
        self.assertEqual({"fake": "data"}, c.data[0])
        self.assertEqual("entanglement_v1", c.quantum_accelerator_module)

        def simple_callable():
            "simple means it doesn't take arguments"
            return "entanglement_v2"

        # TODO - standalone is only calling the callable after _prepare_connection
        # this isn't right
        # c = Connect(engine_url="fake://MyDataset", quantum_accelerator_module=simple_callable)
        # self.assertEqual({"fake": "data"}, c.data[0])

        class QuatumSort(AbstractFakeModel):
            source = Connect(
                engine_url="fake://MyDataset", quantum_accelerator_module=simple_callable
            )

        m1 = QuatumSort()
        self.assertEqual("entanglement_v2", m1.source.quantum_accelerator_module)

    def test_connect_preserve_callable_kwargs(self):
        """
        :class:`ayeaye.connectots.fake.FakeDataConnector` has an optional kwarg-
        'quantum_accelerator_module' set this using a literal or a callable.
        The 'quantum_factory' optional arg shouldn't be called by ayeaye.Connect and should still
        be a callable when used to initiate the FakeDataConnector.
        """

        def q_fact(my_thing):
            return "hello " + str(my_thing)

        class QuatumSort(AbstractFakeModel):
            source = Connect(engine_url="fake://MyDataset", quantum_factory=q_fact)

            def calculate_result(self):
                return self.source.quantum_factory("quantum dynamics")

        m1 = QuatumSort()
        self.assertTrue(callable(m1.source.quantum_factory))
        self.assertEqual("hello quantum dynamics", m1.calculate_result())

    def test_method_overlay(self):
        """
        Add methods to a connector at runtime without inheritance.
        """

        def field_name_aliases(self):
            """
            vanity iterator to alias the 'common_name' field to 'animal_name'
            """
            for r in self:
                r.animal_name = r.common_name
                yield r

        class AnimalsModel(AbstractFakeModel):
            animals = Connect(
                engine_url="csv://" + EXAMPLE_CSV_PATH,
                method_overlay=field_name_aliases,
            )

        m = AnimalsModel()

        # note the use of `field_name_aliases` and `animal_name`
        all_the_animals = [animal.animal_name for animal in m.animals.field_name_aliases()]

        expected = ["Crown of thorns starfish", "Golden dart frog"]
        self.assertEqual(expected, all_the_animals)



try:
    from google.cloud import bigquery
    from google.cloud.exceptions import NotFound
except ModuleNotFoundError:
    pass

from ayeaye.connectors.base import AccessMode, DataConnector


class BigQueryConnector(DataConnector):
    engine_type = "bigquery://"

    optional_args = {"credentials": None, "auto_schema": False}

    def __init__(self, *args, **kwargs):
        """
        Connector to Google Big Query database.
        Args: @see :class:`connectors.base.DataConnector`

        additional args for BigQueryConnector
         'credentials' : (dict) for access within Google Cloud Platform
         'auto_schema' : (bool) only needed for streaming data (e.g. :method:`add`) and only
                         when table doesn't already exist.

        Connection information-
            engine_url format is bigquery://projectId=<projectId>;datasetId=<datasetId>;[tableId=<table>;]
                    note that table is optional.
                    e.g. bigquery://projectId=my_project;datasetId=nice_food;table=cakes;

            format inspired by jdbc:bigquery://[Host]:[Port];ProjectId=[Project];OAuthType= [AuthValue];\
                                                        [Property1]=[Value1];[Property2]=[Value2];
            but it doesn't specify most of the variables in
            https://cloud.google.com/bigquery/docs/reference/rest/

        The bigquery API is asynchronous by way of jobs. This connector doesn't expose async functionality
        but instead waits for jobs to complete. This is purely to keep use of this connector simple.
        """

        super().__init__(*args, **kwargs)

        # set by :method:`connect`
        self.project_id = self.dataset_id = self.table_id = None
        self.client = None

        # these are loaded and created on demand, use via corresponding methods
        self._dataset = self._table_ref = None

        # other
        self.write_buffer_len = 1000  # optimum value not tested for, just guessed
        self.write_rows_buffer = []
        self.table_connection = None  # different from _table_ref, loaded when needed

    def connect(self):
        if self.client is None:
            self.project_id, self.dataset_id, self.table_id = self._decode_engine_url()
            self.client = bigquery.Client(project=self.project_id, credentials=self.credentials)

    def _decode_engine_url(self):
        """
        Returns:
            project, dataset, table : all are str
        """
        r = dict(projectId=None, datasetId=None, tableId=None)
        for param_section in self.engine_url[len(self.__class__.engine_type) :].split(";"):
            if len(param_section) == 0:
                continue
            k, v = param_section.split("=", 1)
            if k in r:
                r[k] = v
        return r["projectId"], r["datasetId"], r["tableId"]

    def write_truncate_file(self, ndjson_fh):
        """
        Send the contents of a new line delimited JSON file to a bigquery table.

        Args:
            ndjson_fh (handle in fileIO type object)
        """
        self.connect()
        job = self.client.load_table_from_file(
            file_obj=ndjson_fh,
            destination=self.table_ref,
            job_config=self._get_ndjson_load_job_config(append_mode=False),
        )
        # wait for it to finish loading
        job.result()

    @property
    def table_ref(self):
        """
        This is the table ref.
        https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.table.TableReference.html
        """
        if not self._table_ref:
            self._table_ref = self.dataset.table(self.table_id)
        return self._table_ref

    @property
    def dataset(self):
        if not self._dataset:
            dataset_ref = self.client.dataset(self.dataset_id)
            try:
                dataset = self.client.get_dataset(dataset_ref)
            except NotFound:
                if self.access != AccessMode.WRITE:
                    raise
                dataset = bigquery.Dataset(dataset_ref)
                dataset.location = "EU"
                dataset = self.client.create_dataset(dataset)
            self._dataset = dataset
        return self._dataset

    def _get_ndjson_load_job_config(self, append_mode=True):
        """
        An instance of this class might create many jobs.

        Args:
            append_mode (bool) if False truncate the table instead
        """
        job_config = bigquery.LoadJobConfig()
        job_config.autodetect = True
        job_config.source_format = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON

        if append_mode:
            job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND
        else:
            job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE

        job_config.ignore_unknown_values = True
        job_config.max_bad_records = 1000000

        return job_config

    @property
    def data(self):
        """
        All the data. Use :method:`query`.
        TODO slices
        """
        self.connect()
        full_qual_table = f"{self.project_id}.{self.dataset_id}.{self.table_id}"
        yield from self.client.list_rows(full_qual_table)

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def add(self, row):
        """
        Add a row to the connected table.
        This method doesn't write to BQ immediately. call :method:`flush_writes` when
        needed and after last row has been written.
        Args:
            row (dict) or tuple with correct column mappings
        """
        self.write_rows_buffer.append(row)
        if len(self.write_rows_buffer) >= self.write_buffer_len:
            self.flush_writes()

    def flush_writes(self):
        """
        Write any buffered data to BigQuery.

        Note that streaming data is treated slightly differently than file loads.
        See
        https://cloud.google.com/blog/products/gcp/life-of-a-bigquery-streaming-insert

        TLDR;
        The streamed data is committed to column stores later. See the 'Streaming buffer
        statistics' section of the table's details section.
        These data does show up in queries but not preview.
        """
        if len(self.write_rows_buffer) == 0:
            # nothing to do
            return

        self.connect()

        if self.table_connection is None:
            # create when needed

            try:
                self.table_connection = self.client.get_table(self.table_ref)
            except NotFound:
                if not self.auto_schema:
                    msg = "Can't write data as table doesn't exist and auto_schema not set."
                    raise ValueError(msg)

                auto_schema = self._auto_schema(self.write_rows_buffer)
                table = bigquery.Table(self.table_ref, schema=auto_schema)
                self.table_connection = self.client.create_table(table)

        errors = self.client.insert_rows(self.table_connection, self.write_rows_buffer)
        # TODO log the errors
        self.write_rows_buffer = []

    def _auto_schema(self, sample_data):
        """
        Returns:
            big query schema (list of bigquery.SchemaField)
        """
        # super simple for now!
        sample_row = sample_data[0]
        schema = []
        for k, v in sample_row.items():
            field_type = "STRING"
            if isinstance(v, int):
                field_type = "INTEGER"
            schema.append(bigquery.SchemaField(k, field_type, mode="REQUIRED"))

        return schema

    def query(self, **kwargs):
        """
        get a subset of :method:`data`.

        Allowed options-
        sql : (str) SQL statement.
        sql_params : (list of tuples) of form
                e.g. [("corpus", "STRING", "romeoandjuliet"),
                      ("min_word_count", "INT64", 250),
                      ]
                      Variable types are those passed to bigquery.ScalarQueryParameter.
                      See https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.query.ScalarQueryParameter.html

                      example
                      sql = 'SELECT * FROM `bbc-datalab.ayeaye_test.my_data` WHERE id > @min_id'
                      sql_params = [("min_id", "INT64", 1234),]

        At present, `sql` arg is mandatory but wont be once we have alternative access patterns.
        """
        assert self.access == AccessMode.READ or AccessMode.READ in self.access

        sql = kwargs.pop("sql", None)
        sql_params = kwargs.pop("sql_params", None)

        if not sql:
            raise NotImplementedError("Only 'sql' has been implemented.")

        self.connect()
        job_config = bigquery.QueryJobConfig()
        if sql_params:
            job_config.query_parameters = [bigquery.ScalarQueryParameter(*sp) for sp in sql_params]
        query_job = self.client.query(sql, job_config=job_config)
        yield from query_job



"""
Run :class:`ayeaye.PartitionedModel` models across multiple operating system processes.
"""
from multiprocessing import Process, Queue

from ayeaye.connect_resolve import connector_resolver


class ProcessPool:
    """
    Like :class:`multiprocessing.Pool` but with instances that persist for as long as the process.

    A worker :class:`multiprocessing.Process` runs methods of an :class:`ayeaye.Model`s instance.
    The model instance is instantiated when the process starts, from there on just the methods
    requested to run the sub-task are called.

    It's frustrating that this can't be done with :class:`multiprocessing.Pool`. Please let me
    know if you can see a way with `Pool`.
    """

    def __init__(self, processes=None, model_initialise=None, context_kwargs=None):
        """
        Indicate how many worker processes to create with either-

        @param processes: (int)
            number of worker processes

            OR

        @param model_initialise (list of args (list) or kwargs (dict))
            Each item in this list is passed as either *args or **kwargs to the Aye-aye model's
            :method:`partition_initialise`.

        @param context_kwargs: (dict)
            connector_resolver context key value pairs.
            The connector resolver is a 'globally accessible' object use to resolve engine_urls.
            It takes key value pairs and callables. This argument is just for the former.
            @see :class:`ayeaye.ayeaye.connect_resolve.ConnectorResolver`
        """
        assert (processes is None) != (model_initialise is None), "Mutually exclusive arguments."

        self.processes = processes
        self.model_initialise = model_initialise
        self.context_kwargs = context_kwargs

    @staticmethod
    def run_model(
        worker_id,
        total_workers,
        ayeaye_model_cls,
        subtask_kwargs_queue,
        return_values_queue,
        initialise,
        context_kwargs,
    ):
        """
        @param worker_id: (int)
            unique number assigned in ascending order to workers as they start

        @param total_workers: (int)
            Number of workers in pool or None for dynamic workers

        @param ayeaye_model_cls: subclass of :class:`ayeaye.PartitionedModel`
            Class, not object/instance.
            This will be instantiated without arguments and subtasks will be methods executed on
            this instance.

        @param subtask_kwargs_queue: :class:`multiprocessing.Queue` object
            subtasks are defined by the (method_name, kwargs) (str, dict) items read from this queue

        @param return_values_queue: :class:`multiprocessing.Queue` object
            method_name, method_kwargs, subtask_return_value from running are sent back to the
            calling the subtask along this queue.

        @param initialise: None, dict or list
            args or kwargs for Aye-aye model's :method:`partition_initialise`

        @param context_kwargs: (dict)
            Output from :method:`connect_resolve.ConnectorResolver.capture_context` - so needs the
            'mapper' key with dict. as value.
            key/values that are made available to the model. These are likely to be taken from
            `self` and passed to this static method as it alone runs in the :class:`Process`.
            @see :class:`connect_resolve.ConnectorResolver`
        """
        # brutal reset but needs a test to ensure threads don't share

        # Depending on OS's fork() the parent process's memory could be available to this Process. Clear
        # all context so only `context_kwargs` is used.
        # For more detail see unittest TestRuntimeMultiprocess.test_resolver_context_not_inherited
        connector_resolver.brutal_reset()

        with connector_resolver.context(**context_kwargs["mapper"]):
            model = ayeaye_model_cls()

            model.runtime.worker_id = worker_id
            model.runtime.total_workers = total_workers

            init_args = []
            init_kwargs = {}
            if initialise is not None:
                for init_as in initialise:
                    if isinstance(init_as, list):
                        init_args = init_as
                    elif isinstance(init_as, dict):
                        init_kwargs = init_as
                    else:
                        raise ValueError("Unknown initialise variable")

            model.partition_initialise(*init_args, **init_kwargs)

            while True:
                method_name, method_kwargs = subtask_kwargs_queue.get()
                if method_name is None:
                    break

                if method_kwargs is None:
                    method_kwargs = {}

                # TODO - :method:`log` for the worker processes should be connected back to the parent
                # with a queue or pipe and it shouldn't be using stdout

                # TODO - supply the connector_resolver context

                # TODO - handle exceptions

                sub_task_method = getattr(model, method_name)
                subtask_return_value = sub_task_method(**method_kwargs)
                return_values_queue.put((method_name, method_kwargs, subtask_return_value))

            model.close_datasets()

    def run_subtasks(self, model_cls, tasks, initialise):
        """
        Generator yielding (method_name, method_kwargs, subtask_return_value) from completed
        subtasks.

        @param ayeaye_model_cls: subclass of :class:`ayeaye.PartitionedModel`
            Class, not object/instance.
            This will be instantiated without arguments and subtasks will be methods executed on
            this instance.

        @param tasks: list of (method_name, method_kwargs) (str, dict)
            each item defines a subtask to execute in a worker process

        @param initialise: None, or list of tuples (dict and or list)
            args or kwargs for Aye-aye model's :method:`partition_initialise`.
            Each item in this list is used to initialise a worker process.
        """
        subtasks_count = len(tasks)

        # Optionally, a model can pass initialisation variables to workers
        if initialise is None:
            worker_init = [None for _ in range(self.processes)]
        else:
            if len(initialise) != self.processes:
                raise ValueError(
                    "The numeber of worker 'initialise' items doesn't match number of workers."
                )
            worker_init = initialise

        context_kwargs = self.context_kwargs or {}

        subtask_kwargs_queue = Queue()
        return_values_queue = Queue()

        proc_table = []
        for proc_id in range(self.processes):
            proc = Process(
                target=ProcessPool.run_model,
                args=(
                    proc_id,
                    self.processes,
                    model_cls,
                    subtask_kwargs_queue,
                    return_values_queue,
                    worker_init[proc_id],
                    context_kwargs,
                ),
            )
            proc_table.append(proc)

        for proc in proc_table:
            proc.daemon = True
            proc.start()

        for sub_task in tasks:
            subtask_kwargs_queue.put(sub_task)

        # instruct worker process to terminate
        for _ in range(self.processes):
            subtask_kwargs_queue.put((None, None))

        for _ in range(subtasks_count):
            yield return_values_queue.get()

        for proc in proc_table:
            proc.join()



"""
Created on 14 Jan 2020

@author: si
"""
import copy
import csv
import os

from ayeaye.connectors.base import AccessMode, FileBasedConnector, FilesystemEnginePatternMixin
from ayeaye.pinnate import Pinnate


class CsvConnector(FileBasedConnector, FilesystemEnginePatternMixin):
    engine_type = "csv://"
    optional_args = {
        "field_names": None,
        "required_fields": None,
        "expected_fields": None,
        "alias_fields": None,
    }
    optional_engine_url_args = FileBasedConnector.optional_engine_url_args + ["start", "end"]
    default_character_encoding = "utf-8-sig"
    write_mode_open_args = {"newline": "\n"}

    def __init__(self, *args, **kwargs):
        """
        Connector to Comma Separated Value (CSV) files.

        For args: @see :class:`connectors.base.FileBasedConnector`

        additional args for CsvConnector
            field_names (sequence, probably a list (str) - Field names for all rows in file.
                    Using this argument when reading forces the CSV module to treat the first line
                    of the file as data, not as a header. When used in write mode it is the order
                    of fields in the output csv file and the :method:`add` can be given a record
                    with additional fields (i.e. not in the `field_names` list) that will be
                    silently ignored and not written to the output file.

            required_fields (sequence, set or list (str)) - fields must be present in header. Other
                    fields may also exist. Raises ValueError if not.
                    Error to call this in write mode. A ValueError will be raised.

            expected_fields (sequence (str)) - fields must be exactly present in header in the
                    order given. Other fields may NOT also exist. Raises ValueError if not.
                    Error to call this in write mode. A ValueError will be raised.

            alias_fields (sequence of str) or (dict)) - if sequence (i.e. a list) it must be the
                    same length as the actual fields in the header. If a dictionary the key is the
                    original field name and the value is alias for the field.
                    Error to call this in write mode or with 'fieldnames' argument. A ValueError
                    will be raised.

        Connection information-
            engine_url format is
            csv://<filesystem absolute path>data_file.csv[;start=<line number>][;end=<line number>][;encoding=<character encoding>]
        e.g. csv:///data/my_project/all_the_data.csv
        """
        self._reset()
        super().__init__(*args, **kwargs)

        # fieldnames are loaded from construction args or from field. This will be unified when
        # schemas are implemented. For now, keep track so loading fieldnames from file doesn't
        # make a :method:`_reset`
        self.base_field_names = copy.copy(self.field_names)
        self.delimiter = ","

        if self.access == AccessMode.READWRITE:
            raise NotImplementedError("Read+Write access not yet implemented")

        self._reset()

    def _reset(self):
        FileBasedConnector._reset(self)
        self.csv = None
        self.approx_position = 0
        if hasattr(self, "base_field_names"):
            self.field_names = copy.copy(self.base_field_names)

    @property
    def engine_params(self):
        """
        @return: (Pinnate) with .file_path
                        and optional: .encoding, .start, .end
        """
        if self._engine_params is None:
            self._engine_params = self._build_engine_params()

            for typed_param in ["start", "end"]:
                if typed_param in self._engine_params:
                    self._engine_params[typed_param] = int(self._engine_params[typed_param])

            if "start" in self._engine_params or "end" in self._engine_params:
                raise NotImplementedError("TODO")

        return self._engine_params

    def connect(self):
        if self.csv is None:
            if self.access == AccessMode.READ:
                if self.field_names is not None and self.alias_fields is not None:
                    msg = "Can't set field_names and alias them. Just use 'field_names'"
                    raise ValueError(msg)

                if self.field_names is not None:
                    extra_args = {"fieldnames": self.field_names}
                else:
                    extra_args = {}

                FileBasedConnector.connect(self)

                self.csv = csv.DictReader(
                    self._file_handle,
                    delimiter=self.delimiter,
                    **extra_args,
                )
                self.field_names = self.csv.fieldnames

                if self.required_fields is not None:
                    required = set(self.required_fields)
                    field_names = set(self.field_names)

                    if not required.issubset(field_names):
                        missing_fields = ",".join(list(required - field_names))
                        msg = f"Missing required field(s): {missing_fields}"
                        raise ValueError(msg)

                if self.expected_fields is not None and self.expected_fields != self.field_names:
                    diff_s = set(self.expected_fields).symmetric_difference(set(self.field_names))
                    diff = ",".join(diff_s)
                    diff_count = len(diff_s)
                    expected = ",".join(self.expected_fields)
                    actual = ",".join(self.field_names)
                    msg = (
                        f"Expected fields does match fields found in file. There are {diff_count} "
                        f"difference(s): [{diff}] expected: [{expected}] but found: [{actual}]"
                    )
                    raise ValueError(msg)

                if self.alias_fields is not None:
                    if isinstance(self.alias_fields, dict):
                        replace_fields = [self.alias_fields.get(f, f) for f in self.csv.fieldnames]
                        self.csv.fieldnames = self.field_names = replace_fields

                    elif not isinstance(self.alias_fields, list) or len(self.alias_fields) != len(
                        self.csv.fieldnames
                    ):
                        msg = (
                            "Alias fields must be a dictionary or list with same number of "
                            "items as fields in the file"
                        )
                        raise ValueError(msg)

                    else:
                        self.csv.fieldnames = self.field_names = self.alias_fields

            elif self.access == AccessMode.WRITE:
                if (
                    self.required_fields is not None
                    or self.expected_fields is not None
                    or self.alias_fields is not None
                ):
                    msg = (
                        "The optional arguments: 'required_fields', 'expected_fields', 'alias_fields' "
                        "can't be used in WRITE mode."
                    )
                    raise ValueError(msg)

                # auto create directory - could be relative path in current working directory
                file_dir = os.path.dirname(self.engine_params.file_path)
                if file_dir and not os.path.exists(file_dir):
                    os.makedirs(file_dir)

                FileBasedConnector.connect(self)

                self.csv = csv.DictWriter(
                    self._file_handle,
                    delimiter=self.delimiter,
                    fieldnames=self.field_names,
                )
                self.csv.writeheader()

            else:
                raise ValueError("Unsupported access mode")

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        self.connect()
        for raw in self.csv:
            # OSError: telling position disabled by next() call so this for now
            # str(x) will slightly over count 'None'. None is given by DictReader when
            # trailing commas are omitted for optional fields at end of row.
            self.approx_position += len(self.delimiter.join([str(x) for x in raw.values()]))
            yield Pinnate(data=raw)

    @property
    def data(self):
        raise NotImplementedError("TODO")

    def add(self, data):
        """
        Write line to CSV file.
        @param data: (dict or Pinnate)
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        # until schemas are implemented, first row determines fields
        if self.csv is None and self.field_names is None:
            if isinstance(data, dict):
                self.field_names = list(data.keys())
            elif isinstance(data, Pinnate):
                self.field_names = list(data.as_dict().keys())

        self.connect()

        if isinstance(data, dict):
            _d = data
        elif isinstance(data, Pinnate):
            _d = data.as_dict()
        else:
            raise ValueError("data isn't an accepted type. Only (dict) or (Pinnate) are accepted.")

        if self.field_names:
            # the CSV module needs the dictionary passed to write row to match the fieldnames. It's
            # a common scenario to extract just a few fields, these have already been passed to
            # the CsvConnector so just extract the fields needed.
            data_extract = {fn: _d[fn] for fn in self.field_names if fn in _d}
        else:
            data_extract = _d

        self.csv.writerow(data_extract)


class TsvConnector(CsvConnector):
    """
    Tab separated values. See :class:`CsvConnector`
    """

    engine_type = "tsv://"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # see note in CsvConnector
        self.base_field_names = copy.copy(self.field_names)
        self.delimiter = "\t"



from ayeaye.common_pattern.manifest import AbstractManifestMapper, EngineFromManifest
from ayeaye.connect_resolve import connector_resolver
from ayeaye.connect import Connect
from ayeaye.connectors import connector_registry
from ayeaye.connectors.base import AccessMode
from ayeaye.model import LockingMode, Model, PartitionedModel
from ayeaye.model_collection import ModelCollection
from ayeaye.pinnate import Pinnate



from inspect import isclass

from .base import DataConnector
from .bigquery import BigQueryConnector
from .csv_connector import CsvConnector, TsvConnector
from .elasticsearch_connector import ElasticsearchConnector
from .fake import FakeDataConnector
from .json_connector import JsonConnector
from .kafka_connector import KafkaConnector
from .ndjson_connector import NdjsonConnector
from .parquet_connector import ParquetConnector
from .restful_connector import RestfulConnector
from .sqlalchemy_database import SqlAlchemyDatabaseConnector
from .uncooked_connector import UncookedConnector

from ayeaye.ignition import Ignition, EngineUrlCase, EngineUrlStatus


class ConnectorPluginsRegistry:
    def __init__(self):
        self.registered_connectors = []  # list of classes, not instances - publicly readable
        self.reset()

    def reset(self):
        "set registered connectors to just those that are built in"
        self.registered_connectors = [
            BigQueryConnector,
            CsvConnector,
            FakeDataConnector,
            KafkaConnector,
            ParquetConnector,
            TsvConnector,
            SqlAlchemyDatabaseConnector,
            JsonConnector,
            NdjsonConnector,
            ElasticsearchConnector,
            UncookedConnector,
            RestfulConnector,
        ]

    def register_connector(self, connector_cls):
        """
        @param connector_cls ():
        """
        if not isclass(connector_cls) or not issubclass(connector_cls, DataConnector):
            msg = "'connector_cls' should be a class (not object) and subclass of DataConnector"
            raise TypeError(msg)

        # MAYBE - a mechanism to specify the position/priority of the class
        self.registered_connectors.append(connector_cls)


# global registry
connector_registry = ConnectorPluginsRegistry()


def connector_factory(engine_url):
    """
    return a subclass of DataConnector
    @param engine_url (str):
    """
    if isinstance(engine_url, str) and "://" not in engine_url:
        # The engine type is only available after the context has been resolved. Don't error here
        # if resolution isn't yet possible.
        ignition = Ignition(engine_url)
        try:
            status, e_url = ignition.engine_url_at_state(EngineUrlCase.FULLY_RESOLVED)
            if status == EngineUrlStatus.OK:
                engine_url = e_url
        except ValueError:
            # Full resolve not available yet. It could be possible to use a partially
            # resolved url if this current behaviour isn't good enough.
            pass

    engine_type = engine_url.split("://", 1)[0] + "://"
    for connector_cls in connector_registry.registered_connectors:
        if isinstance(connector_cls.engine_type, list):
            supported_engines = connector_cls.engine_type
        else:
            supported_engines = [connector_cls.engine_type]

        if engine_type in supported_engines:
            return connector_cls

    raise NotImplementedError(f"Unknown engine in url:{engine_url}")



"""
Created on 19 May 2021

@author: si
"""
from ayeaye.connectors.base import AccessMode, FileBasedConnector, FilesystemEnginePatternMixin


class UncookedConnector(FileBasedConnector, FilesystemEnginePatternMixin):
    engine_type = "file://"
    optional_args = {
        "file_mode": "t",
    }

    def __init__(self, *args, **kwargs):
        """
        Connector to use a raw local file.

        The file will be opened on demand. i.e. reading .data or getting the .file_handle. Getting
        the .file_path will not open the file.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for UncookedConnector
            file_mode (str) - Either 'b' for binary or 't' (default) text. Other modes not yet
                    supported.
                    Mode to open the file in. See-
                    https://docs.python.org/3/library/functions.html#open

        Connection information-
            engine_url format is
            file://<filesystem absolute path>[;encoding=<character encoding>]
        e.g. file:///data/my_project/interesting_notes.txt;encoding=latin-1
        """
        self._reset()
        super().__init__(*args, **kwargs)

        if self.access == AccessMode.READWRITE:
            raise NotImplementedError("READWRITE access not yet implemented")

        # this overwrites a class variable in :class:`FileBasedConnector`
        if self.file_mode not in ["b", "t"]:
            raise ValueError(f"File mode: {self.file_mode} not supported")

    def _reset(self):
        FileBasedConnector._reset(self)
        self._file_content = None  # used in read mode

    @property
    def file_handle(self):
        """
        File handle to open file for operations such as :method:`read`()
        """
        self.connect()
        return self._file_handle

    def __len__(self):
        """
        @return: (int) file size
        """
        if self.access != AccessMode.READ:
            raise NotImplementedError("Not yet available in write mode")

        self.connect()
        return self.file_size

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        "Iteration not possible"

        # MAYBE make a readlines method?
        raise TypeError("Can't iterate through a file")

    @property
    def data(self):
        """
        @return: contents of entire file. Type depends on encoding arguments.
        """
        if self.access != AccessMode.READ:
            raise ValueError("Not open in read mode")

        if self._file_content is None:
            self._file_content = self.file_handle.read()

        return self._file_content

    @data.setter
    def data(self, file_content):
        """
        attribute based setter.

        e.g.
        c = UncookedConnector(engine_url="file://mydata_file", access=ayeaye.AccessMode.WRITE)
        c.data = "hello world!"
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Not open in write mode")

        file_content = self.file_handle.write(file_content)

    @property
    def schema(self):
        raise TypeError("Uncooked connectors don't have rich interfaces like schemas")



"""
A manifest is "a list of people and goods carried on a ship or plane" but here, it's a list of
datasets.

A common pattern is to bootstrap a batched build by listing all the target files used in the build in a
manifest file. This manifest file is a dataset for any model that works on the target files. 
"""
import copy

from collections import defaultdict
from inspect import ismethod

import ayeaye


class EngineFromManifest:
    """
    Make engine_urls stored in a manifest file available to :class:`ayeaye.Connector`s in an
    :class:`ayeaye.Model`.

    Each engine_url represents a specific version of a source dataset so this pattern can be
    used to create versioned ETL builds by storing versioning information in a manifest file.

    A manifest file is a dataset which can be retained to re-create ETL builds at a later date
    and to show data provenance.

    Creating the manifest file itself isn't expected to be a deterministic process as it probably
    takes a filesystem listing of available datasets and from this determines the most recent
    versions to use in the manifest file.

    Use a subclass of :class:`DataConnector` that supports dictionary access to `.data` as a
    manifest. Use this to build engine_url(s) for passing to another dataset within the
    dataset declarations part of an :class:`ayeaye.Model`.

    Tip: JSON is a good choice for the manifest dataset.

    example-

    The file `products_manifest_20210618.json` contains

    ```
    {"source_files": "products_inventory_2021.csv"}`
    ```

    ```
    import ayeaye
    from ayeaye.common_pattern.manifest import EngineFromManifest
    from ayeaye.connect_resolve import connector_resolver

    class ProductsBuild(ayeaye.Model):
        manifest = ayeaye.Connect(engine_url="json://products_manifest_{build_id}.json")
        products = ayeaye.Connect(engine_url=EngineFromManifest(manifest, "source_files", "csv"))


    with connector_resolver.context(build_id="20210618"):
        m = ProductsBuild()
        assert m.products.engine_url == "csv://products_inventory_2021.csv"
    ```
    """

    def __init__(self, manifest_dataset, field_name, engine_type):
        """
        :param manifest_dataset (subclass :class:`DataConnector` object): with .data and dictionary
                access to .data.
        :param field_name (str): field within manifest_dataset.data[field_name]
        :param engine_type (str): prefix to engine_url. e.g. 'json' would give 'json://'
        """
        self.manifest_dataset = manifest_dataset

        self.field_name = field_name
        self.engine_type = engine_type

    def __call__(self):
        """
        When used according to the pattern above this will be called on demand by
        :class:`ayeaye.Connect` when a dataset is accessed so the global connector_resolver will
        have been setup with any context that is needed before reading the manifest_dataset. That
        context wouldn't have been available during construction, hense a callable.
        """
        if isinstance(self.manifest_dataset, ayeaye.Connect):
            # .clone() is to prevent the .Connect being bound to the parent if it's connected
            manifest_dataset = self.manifest_dataset.clone()
        else:
            manifest_dataset = self.manifest_dataset

        # create ephemeral dataset not tied to an ayeaye.Model
        e_url = ayeaye.connector_resolver.resolve(manifest_dataset.engine_url)
        ds = ayeaye.Connect(engine_url=e_url)
        dataset_section = ds.data[self.field_name]

        # :class:`connectors.multi_connector.MultiConnector` vs. single engine_urls connectors
        if isinstance(dataset_section, list):
            return [f"{self.engine_type}://{f}" for f in dataset_section]

        return f"{self.engine_type}://{dataset_section}"


class AbstractManifestMapper:
    """
    A manifest is a dataset containing a list of other datasets.

    It's common for an ETL model to transform or map each of these source datasets into a corresponding
    output dataset(s).

    :class:`AbstractManifestMapper` should be subclassed to act as a mapper between datasets in a
    manifest and output datasets. It is used to declare data connections within an :class:`ayeaye.Model`
    (i.e. the :class`ayeaye.Connect` class variables) and it's used within the model to provide
    maps of input to output datasets.

    For the manifest dataset, use a subclass of :class:`DataConnector` that supports dictionary
    access to `.data`. JSON is a good choice for the manifest dataset.

    Sub classes implement a method for each map and this method is either a generator or returns
    a list of (original_manifest_item, engine_url). [Note - this doesn't strictly need to be an
    engine_url if the mapper is used more generally and not just as a helper for ayeaye.Connect
    as is being demonstrated here].

    For example, using xxxx as the map name:

    ```
    class FileMapper(AbstractManifestMapper):
        def map_xxxx(self):
            return [(f, f"json://{f}") for f in self.manifest_items]
    ```

    creates a method `.xxxx()` which returns just the engine_urls. It is used in an ayeaye.Connect-

    ```
    class AustralianAnimals(ayeaye.Model):
        animals_manifest = ayeaye.Connect(engine_url="json://{input_data}/animals_manifest.json")
        animals_mapper = FileMapper(animals_manifest)
        all_files = ayeaye.Connect(engine_url=animals_mapper.all_files)
    ```

    And FileMapper will also be an iterator returning a object with attributes linking to the
    engine_urls of all mappers.

    For an example of how to use this see-
    https://github.com/Aye-Aye-Dev/AyeAye/blob/master/examples/manifest_mapper.py

    The tests in :class:`tests.common_pattern.test_manifest.TestManifest` also demonstrate usage.

    Notes-

    This class is a descriptor as it's expected to be used to declare class variables that are then
    used within the instance.

    If the subclass changes the constructor's arguments you must implement the __copy__ method.
    """

    def __init__(self, manifest_dataset, field_name=None):
        """
        :param manifest_dataset (subclass :class:`DataConnector` object): with .data and dictionary
                access to .data.
        :param field_name (str): field within manifest_dataset.data[field_name]
        """
        self.manifest_dataset_unresolved = manifest_dataset
        self.field_name = field_name

    def __copy__(self):
        c = self.__class__(manifest_dataset=self.manifest_dataset_unresolved, field_name=self.field_name)
        return c

    def __get__(self, instance, instance_class):
        if instance is None:
            # class method called.
            # This means `self` is currently an attribute of the class (so NOT an instance
            # variable).
            #
            # see https://docs.python.org/3/howto/descriptor.html
            return self

        ident = id(self)
        # dynamically create a place in the parent class to hold descriptor objects
        if not hasattr(instance, "_abstract_manifes_mapper_vars"):
            instance._abstract_manifes_mapper_vars = {}

        if ident not in instance._abstract_manifes_mapper_vars:
            instance._abstract_manifes_mapper_vars[ident] = copy.copy(self)

        return instance._abstract_manifes_mapper_vars[ident]

    @property
    def manifest_items(self):
        """
        This will be called on demand after the global connector_resolver has been setup with any
        context that is needed before reading the manifest_dataset. That context wouldn't have been
        available during construction.

        Generator yielding each item in the manifest datasets's target field.
        i.e. list of items in manifest.data[self.field_name]
        """
        if isinstance(self.manifest_dataset_unresolved, ayeaye.Connect):
            # .clone() is to prevent the .Connect being bound to the parent if it's connected
            manifest_dataset = self.manifest_dataset_unresolved.clone()
        else:
            manifest_dataset = self.manifest_dataset_unresolved

        # create ephemeral dataset not tied to an ayeaye.Model
        e_url = ayeaye.connector_resolver.resolve(manifest_dataset.engine_url)
        self._manifest_dataset = ayeaye.Connect(engine_url=e_url)

        if self.field_name is None:
            yield from self._manifest_dataset.data
        else:
            yield from self._manifest_dataset.data[self.field_name]

    @property
    def manifest_data(self):
        """
        This will be called on demand after the global connector_resolver has been setup with any
        context that is needed before reading the manifest_dataset. That context wouldn't have been
        available during construction.

        Returns content of manifest's data attribute. Similar to :method:`manifest_items`.
        """
        if isinstance(self.manifest_dataset_unresolved, ayeaye.Connect):
            # .clone() is to prevent the .Connect being bound to the parent if it's connected
            manifest_dataset = self.manifest_dataset_unresolved.clone()
        else:
            manifest_dataset = self.manifest_dataset_unresolved

        # create ephemeral dataset not tied to an ayeaye.Model
        e_url = ayeaye.connector_resolver.resolve(manifest_dataset.engine_url)
        self._manifest_dataset = ayeaye.Connect(engine_url=e_url)

        if self.field_name is None:
            return self._manifest_dataset.data
        else:
            return self._manifest_dataset.data[self.field_name]

    @property
    def methods_mapper(self):
        """
        mapper methods are those with names starting `map_`.

        The `map_name` is the part afer `map_`.

        For example-

        ```
        def map_xxxx(self):
            ...
        ```

        `xxxx` is the `map_name`.

        mapper methods return a list of tuples.

        @returns (dict) map_name -> mapping_method bound to instance
        """
        method_prefix = "map_"
        _methods = {}
        for obj_name in dir(self):

            if not obj_name.startswith(method_prefix):
                continue

            obj = getattr(self, obj_name)
            if ismethod(obj):
                map_name = obj_name[len(method_prefix) :]

                if map_name == "manifest_item":
                    raise ValueError("Reserved method name: manifest_item")

                _methods[map_name] = obj

        return _methods

    @property
    def full_map(self):
        """

        manifest_item -> map_name -> [mapped_value[,mapped_value...]]

        """
        # loop through all items in manifest. This ensures at least
        # an empty map_name dictionary if no mapper targets that item.
        full_map = {}
        for manifest_listed_file in self.manifest_items:
            full_map[manifest_listed_file] = defaultdict(list)

        for map_name, map_method in self.methods_mapper.items():

            for manifest_listed_file, engine_url in map_method():
                full_map[manifest_listed_file][map_name].append(engine_url)

        return full_map

    def __iter__(self):
        """
        yield :class:`ayeaye.Pinnate` objects with map_names as the attributes and the additional
        attribute `manifest_item`.
        """
        for manifest_item, map_values in self.full_map.items():
            m = {"manifest_item": manifest_item}
            for map_name, engine_urls in map_values.items():

                # 1-1 mapping easier to understand when returned as single item
                # if this isn't intuitive then map_method() should declare what to do here.
                # e.g.
                # always return list (possibly with one item) when dealing with multi-connector.
                m[map_name] = engine_urls[0] if len(engine_urls) == 1 else engine_urls

            yield ayeaye.Pinnate(m)

    def __getattr__(self, attr):
        """
        map_xxx() method becomes .xxx() method and .xxx() only returns the engine_urls, not the file
        from the manifest. These `map_name` methods are for passing engine_urls to
        :class`ayeaye.Connect` in the class variable dataset declaration part of an
        :class:`ayeaye.Model` where the mapper will be called later.
        """
        if attr not in self.methods_mapper:
            cls_name = self.__class__.__name__
            attrib_error_msg = f"'{cls_name}' object has no attribute '{attr}'"
            raise AttributeError(attrib_error_msg)

        full_map_name_method = self.methods_mapper[attr]

        def engine_url_reducer():
            return [m[1] for m in full_map_name_method()]

        return engine_url_reducer



from enum import Enum
import glob
import os
import types

from ayeaye.ignition import Ignition, EngineUrlCase, EngineUrlStatus


class AccessMode(Enum):
    READ = "r"
    WRITE = "w"
    READWRITE = "rw"


class AbstractExpandEnginePattern:
    """
    Mixin to support wildcard engine_urls.
    https://www.pythontutorial.net/python-oop/python-mixin/

    When an engine_url contains pattern matching characters the :class:`MultiConnector` should
    be used after the :class:`ConnectorResolver` singleton has resolved context variables.

    Subclasses of this abstract class are given responsibility to pattern match as filesystems
    and blob services use different methodologies to do this.
    """

    def has_multi_engine_pattern(self):
        """
        @return: bool
            the engine_url for the class implementing this mixin has a pattern
        """
        raise NotImplementedError("Must be implemented by subclasses")

    def expand_pattern(self):
        """
        @return: list of str
            engine_urls, each one becomes a connector (subclass of :class:`DataConnector`) in a
            :class:`MultiConnector`
        """
        raise NotImplementedError("Must be implemented by subclasses")


class FilesystemEnginePatternMixin(AbstractExpandEnginePattern):
    """
    Pattern match files and directories. @see :class:`AbstractExpandEnginePattern`
    """

    # List of characters that when found in an engine_url indicate it's a pattern matching url that
    # will result in multiple engine_urls. i.e. file:///data/*.csv
    # A :class:`MultiConnector` should be used.
    # Optionally defined by subclasses.
    pattern_characters = ["*", "?"]

    def has_multi_engine_pattern(self):
        for pattern_indicating_character in self.pattern_characters:
            if pattern_indicating_character in self.engine_url:
                return True
        return False

    def expand_pattern(self):
        status, e_url = self.ignition.engine_url_at_state(EngineUrlCase.FULLY_RESOLVED)
        if status != EngineUrlStatus.OK:
            raw_e_url = self.ignition.engine_url_at_state(EngineUrlCase.RAW)
            msg = f"Engine url ({raw_e_url}) couldn't be resolved enough to pattern match."
            raise ValueError(msg)

        # strip engine type
        # for now, they all need to have the same engine_type. Maybe engine_url starts
        # with `://` for auto detect based on file name.
        engine_type, engine_path_pattern = e_url.split("://", 1)

        engine_url = []
        for engine_file in glob.glob(engine_path_pattern):
            engine_url.append(f"{engine_type}://{engine_file}")

        return engine_url


class DataConnector:
    # must be defined by subclasses. Str or list of str of engine_types supported by
    # the connector. e.g. engine_type = "ndjson://"
    engine_type = None

    # wildcards can be expanded by subclasses to match multiple data sources which each become
    # a member of a :class:`MultiConnector`. When set this class will be the default expander.
    # It is instantiated when the subclass of DataConnector is constructed.
    # `engine_pattern_expander` must be a subclass of :class:`AbstractExpandEnginePattern`.
    engine_pattern_expander = None

    # subclasses should specify their optional kwargs. Values in this dict are default values.
    optional_args = {}

    # list/set of arg names that are present in optional_args and when they are assigned with a
    # callable this shouldn't be called by :class:`ayeaye.Connect` but should instead be passed
    # as is to the target subclass of :class:`ayeaye.DataConnector`.
    preserve_callables = []

    # TODO - make it possible for internal variable name to not match kwarg name, e.g. schema -> self._schema
    # TODO - these aren't always optional, handling missing mandatory args here

    def __init__(self, engine_url=None, access=AccessMode.READ, method_overlay=None, **kwargs):
        """
        API to interact with AyeAye-compatible data sources
        :param engine_url (string): The file or URI to the location of the dataset
        :param access (AccessMode): Whether the dataset accessed through this Connector is for
                input or output
        :param method_overlay (callable or list of callables): Callables are patched onto `self` as methods.
                e.g.
                    my_dataset = ayeaye.Connect(..., method_overlay=[accessor_method])
                    ...
                    my_dataset.accessor_method() now exists

        **kwargs are any params needed by subclasses

        Note that subclasses must call this constructor and should 'pop' their arguments so that
        none are left unprocessed. Any which are specified in optional_args will be set as
        attributes by this super constructor.
        """
        # set when :class:`ayeaye.Connect` builds subclass instances
        self._connect_instance = None  # instance of :class:`ayeaye.Connect`
        self._parent_model = None  # instance of :class:`ayeaye.Model`

        self.access = access

        # dynamically add methods to `self`
        if method_overlay:
            overlays = [method_overlay] if callable(method_overlay) else method_overlay
            for m_overlay in overlays:
                method_name = m_overlay.__name__
                setattr(self, method_name, types.MethodType(m_overlay, self))

        # engine_urls may need resolution of templated variables (typically secrets and paths). The
        # :class:`Ignition` module does this and makes URLs croxx-operating system compatible.
        self.ignition = Ignition(engine_url)

        # It might be both possible and necessary to resolve the engine_url. This happens when the
        # engine_type is within the resolve context. e.g. engine_url="{my_engine_type}://somedata"
        try:
            status, e_url = self.ignition.engine_url_at_state(EngineUrlCase.FULLY_RESOLVED)
            if status == EngineUrlStatus.OK:
                engine_url = e_url
        except ValueError:
            # no problem if not resolvable yet
            pass

        if isinstance(engine_url, str):
            engine_type = (
                [self.engine_type] if isinstance(self.engine_type, str) else self.engine_type
            )
            if not any([engine_url.startswith(et) for et in engine_type]):
                raise ValueError("Engine type mismatch")

        # process optional arguments with their defaults
        for arg in self.__class__.optional_args:
            setattr(self, arg, kwargs.pop(arg, self.__class__.optional_args[arg]))

        # Subclasses should consume any kwargs before this constructor is invoked
        if len(kwargs) > 0:
            what_are_these = ", ".join(kwargs.keys())
            raise ValueError(f"Unexpected argument(s): '{what_are_these}'")

    def __del__(self):
        self.close_connection()

    @property
    def connect_instance(self):
        """
        Instances of subclasses of :class:`DataConnector` and :class:`ModelConnector` are usually
        built by :class:`ayeaye.Connect`. `connect_instance` is a reference to make it easy to
        tweak an existing connect on a model. See :method:`TestConnect.test_update_by_replacement`
        for an example.
        """
        return self._connect_instance

    @property
    def engine_url(self):
        """
        A fully resolved engine url contains everything needed to connect to the data source. This
        includes secrets like username and password.

        Don't store the result of this property when locking or similar. @see :class:`Ignition`.

        @return: (str) the fully resolved engine url.
        """
        status, e_url = self.ignition.engine_url_at_state(EngineUrlCase.FULLY_RESOLVED)
        if status == EngineUrlStatus.OK:
            return e_url
        elif status == EngineUrlStatus.NOT_AVAILABLE:
            raise ValueError("Engine URL not available")
        else:
            raise ValueError(f"Engine URL failed to resolve: {status}")

    def update(self, **kwargs):
        """
        Rebuild this (subclass of) DataConnector. It can only be used if self was built by
        :class:`ayeaye.Connect` because Connect determines the subclass based on the engine type.
        If the call used the same engine_type as `self` then this wouldn't be needed.

        The instance of :class:`Connect` that `self` is referencing is updated with `kwargs`. The
        current prepared DataConnect subclass (i.e. self) that was assigned to the parent model is
        thrown away because the new args probably alter how the dataset is connected to. It will
        re-initialise on demand.

        @param kwargs: anything you'd pass to :class:`Connect`
        @return: None
        """
        if self._connect_instance is None:
            raise ValueError("Connect instance not referenced. See :method:`update` for more")

        self._connect_instance.update(**kwargs)
        ident = id(self._connect_instance)
        if self._parent_model is not None:
            # remove old Connector, it will be re-built on access
            del self._parent_model._connections[ident]

    def connect(self):
        """
        Open resource handles used to access the dataset. e.g. network or filesystem connection.
        These resources are help open by the subclass.
        """
        pass

    def close_connection(self):
        """
        Explicitly close the connection to the dataset rather than just wait for the process to end.
        """
        pass

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        """
        more intuative use of the data behind each the DataConnector.
        e.g.
        ...
        for record in self.my_dataset:
            print(record)
        """
        self.connect()
        return self.data

    @property
    def data(self):
        """
        Return the entire dataset. The returned objects could be lazy evaluated. Use of this method
        will get ugly with large datasets. Better to use the :method:`__iter__` iterator.

        @return: mixed
        """
        raise NotImplementedError("TODO")

    def as_pandas(self):
        """
        Similar to :method:`data` but as a Pandas dataframe.

        @return: (Pandas dataframe)
        """
        raise NotImplementedError("Not available for all datasets or might need to be written")

    @property
    def progress(self):
        """
        Return a number between and including 0 to 1 as simplified representation of how much of
        the dataset has been read.

        Optional property that can be implemented by subclasses. When available it returns float
        from 0 -> 1 describing the position within the dataset. The value is only meaningful when
        the dataset is in READ mode and records are being iterated through. The value isn't
        expected to be really accurate as some reads will cache data so exact file positions wont
        be available.

        :returns: (float) or None when not available.
        """
        return None

    @property
    def datasource_exists(self):
        msg = (
            "Can optionally be implemented by subclasses. "
            "Contribute if you need it! There is an example in JsonConnector."
        )
        raise NotImplementedError(msg)


class FileBasedConnector(DataConnector):
    """
    'File based' means the operating system is opening a file handle from a path.

    This is a mixin for common functionality between file based `DataConnector` modules.
    """

    # class variables that can be defined by subclasses or left as are or overwritten with instance
    # variable.
    optional_engine_url_args = ["encoding"]  # list of str
    default_character_encoding = None
    write_mode_open_args = {}
    # files will be opened in text mode
    file_mode = "t"

    def _reset(self):
        """
        Subclasses must call this within the constructor
        """
        self._file_handle = None
        self._encoding = None
        self._engine_params = None
        self.file_size = None

    @property
    def engine_params(self):
        """
        @return: (Pinnate) with .file_path
                        and optional: .encoding
        """
        if self._engine_params is None:
            self._engine_params = self._build_engine_params()

        return self._engine_params

    def _build_engine_params(self):
        """
        can be overridden by subclass.
        @see :method:`engine_params`
        """

        ep = self.ignition._decode_filesystem_engine_url(
            self.engine_url, optional_args=self.optional_engine_url_args
        )

        if "encoding" in ep:
            self._encoding = ep.encoding

        return ep

    def connect(self):
        if self._file_handle is None:
            if self.file_mode == "b" and self.encoding is not None:
                raise ValueError("Binary file mode can't be set with an encoding")

            if self.access == AccessMode.READ:
                file_mode = "r" + self.file_mode
                self._file_handle = open(
                    self.engine_params.file_path, file_mode, encoding=self.encoding
                )
                self.file_size = os.stat(self.engine_params.file_path).st_size

            elif self.access == AccessMode.WRITE:
                file_mode = "w" + self.file_mode
                self._file_handle = open(
                    self.engine_params.file_path,
                    file_mode,
                    encoding=self.encoding,
                    **self.write_mode_open_args,
                )
            elif self.access == AccessMode.READWRITE:
                # this is a tricky mode because of flushes, truncates and opening a file
                # which may or may not exist.

                file_exists = os.path.exists(self.engine_params.file_path)

                if file_exists:
                    file_mode = "r" + self.file_mode + "+"
                else:
                    file_mode = "w" + self.file_mode + "+"

                self._file_handle = open(
                    self.engine_params.file_path,
                    file_mode,
                    encoding=self.encoding,
                )
            else:
                raise ValueError("Unknown access mode")

    @property
    def file_path(self):
        """
        @return: (str) filesystem path to file
        """
        return self.engine_params.file_path

    @property
    def encoding(self):
        """
        default encoding. 'sig' means don't include the unicode BOM
        """
        if self._encoding is None:
            ep = self.engine_params
            self._encoding = ep.encoding if "encoding" in ep else self.default_character_encoding

        return self._encoding

    def close_connection(self):
        if self._file_handle is not None:
            self._file_handle.close()
        self._reset()

    @property
    def progress(self):
        if self.access != AccessMode.READ or self.file_size is None or self.approx_position == 0:
            return None

        return self.approx_position / self.file_size



from inspect import isclass

from .abstract_modifier import AbstractModifier
from .smart_open_modifier import SmartOpenModifier


# def modifier_factory(*modifiers):
#     """
#     Build a chain of instantiated modifiers so the :method:`apply` from the super class
#     `AbstractModifier` can be called once for a :class:`DataConnector` subclass object.
#
#     return a subclass of EngineTypeModifier
#     @param modifiers (list of str):
#     """
#     if len(modifiers) != 1:
#         raise NotImplementedError("Current supports a single engine type modifier")
#
#     raise NotImplementedError("TODO - I am here")
#
#
# def apply_modifier(connector, *modifiers):
#     """
#     maybe??
#     """
#     pass


class ConnectorModifiersPluginsRegistry:
    """
    A modifier adds capabilities to a connector, for example, transparent compression.
    """

    def __init__(self):
        self.registered_modifiers = []  # list of classes, not instances - publicly readable
        self.reset()

    def reset(self):
        "set registered connectors to just those that are built in"
        self.registered_modifiers = [
            SmartOpenModifier,
        ]

    def register_connector(self, modifier_cls):
        """
        Add a private modifier to the dataset connection discovery process.
        @param modifier_cls ():
        """
        if not isclass(connector_cls) or not issubclass(connector_cls, AbstractModifier):
            msg = "'connector_cls' should be a class (not object) and subclass of DataConnector"
            raise TypeError(msg)

        # MAYBE - a mechanism to specify the position/priority of the class
        self.registered_connectors.append(connector_cls)


# global registry
connector_modifiers_registry = None



'''
Created on 16 Jun 2021

@author: si
'''
import os
import unittest

import ayeaye

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_DATA_PATH = os.path.join(PROJECT_TEST_PATH, 'data')


class FakeModel(ayeaye.Model):
    animals = ayeaye.Connect(engine_url=f"csv://{TEST_DATA_PATH}/" + "{animal_type}.csv",
                             )

    def build(self):
        for a in self.animals:
            self.log(a.common_name)


class TestModelLocking(unittest.TestCase):

    def test_simple_mapping(self):
        """
        Dictionary mapping passed to .context should be in the lock document.
        """
        # data_version
        with ayeaye.connector_resolver.context(animal_type='deadly_creatures'):
            m = FakeModel()
            lock_doc = m.lock()

        self.assertEqual('deadly_creatures', lock_doc['resolve_context']['mapper']['animal_type'])

    def test_lock_all_datasets(self):
        """
        With LockingMode.ALL_DATASETS , engine URLS from all datasets are captured
        """
        with ayeaye.connector_resolver.context(animal_type='deadly_creatures'):
            m = FakeModel()
            lock_doc = m.lock(locking_level=ayeaye.LockingMode.ALL_DATASETS)

        expected_file_ending = 'tests/data/deadly_creatures.csv'
        self.assertTrue(lock_doc['dataset_engine_urls']['animals'].endswith(expected_file_ending))



from ayeaye.connectors.base import DataConnector


class PlaceholderDataConnector(DataConnector):
    """
    For use when the engine_type isn't available yet. The DataConnector abstract class is a
    callable that can convert a DataConnector belonging to an :class`ayeaye.Model` into
    another subclass of DataConnector.
    """

    engine_type = None

    def connect(self):
        pass

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    @property
    def schema(self):
        return None

    @property
    def data(self):
        return None



import json


class Pinnate:
    """
    Dictionary or attribute access to variables loaded either from a JSON
    string or supplied as a dictionary.

    >>> a = Pinnate({'my_string':'abcdef'})
    >>> a.my_string
    'abcdef'
    >>> a['my_string']
    'abcdef'
    >>> a.as_dict()
    {'my_string': 'abcdef'}

    objects within lists-

    >>> from ayeaye.pinnate import Pinnate
    >>> d={'my_things' : [1,2,{'three':3}]}
    >>> a = Pinnate(d)
    >>> a.my_things
    [1, 2, <ayeaye.pinnate.Pinnate object at 0x108526e10>]
    >>> a.my_things[2].three
    3
    """

    def __init__(self, data=None):
        """
        :param data: mixed
            can be dictionary or list or set or dictionary/list encoded in json or instance of Pinnate
        """
        # this is the 'payload', it's type is decided on first use. It's typically a dictionary because
        # the attribute nature of Pinnate is the most useful feature. It can also be a list or set.
        self._attr = None

        if isinstance(data, self.__class__):
            self._attr = data._attr
        elif data:
            self.load(data)

    @property
    def payload_undefined(self):
        """
        No data has been set

        @return: boolean
            No data has been provided so _attrib's type hasn't yet been determined
        """
        return self._attr is None

    def is_payload(self, *payload_type):
        """
        :class:`Pinnate` can hold mixed data types. Inspect current payload type.

        e.g.
        >>> p = Pinnate({1:2})
        >>> p.is_payload(dict)
        True

        @param *payload_type: type
            e.g. dict, set or list
            when multiple payload types are given just one has to match

        @return: boolean
        """
        return any([type(self._attr) == pt for pt in payload_type])

    def __unicode__(self):
        as_str = str(self._attr)
        return f"<Pinnate {as_str}>"

    def __str__(self):
        return self.__unicode__().encode("ascii", "replace").decode()

    def keys(self):
        if self.payload_undefined or not self.is_payload(dict):
            raise TypeError("Payload data isn't a dictionary")

        return self._attr.keys()

    def values(self):
        if self.payload_undefined:
            raise TypeError("Payload data hasn't been set")

        if self.is_payload(dict):
            return self._attr.values()

        if self.is_payload(list, set):
            return self._attr

        raise TypeError("Unknown payload data type")

    def items(self):
        if self.payload_undefined or not self.is_payload(dict):
            raise TypeError("Payload data isn't a dictionary")

        return self._attr.items()

    def __contains__(self, key):
        if self.payload_undefined:
            return False

        if self.is_payload(dict, set):
            return key in self._attr

        raise TypeError("Operation not possible with current payload data type")

    def __iter__(self):
        """
        return a generator

        For lists and sets the generator yields each item. For dictionaries it yield (key, value)
        """
        if self.is_payload(set, list):
            return iter(self._attr)

        if self.is_payload(dict):
            as_key_pairs = [(k, v) for k, v in self._attr.items()]
            return iter(as_key_pairs)

    def as_dict(self, select_fields=None):
        """
        @param select_fields: (list of str) to only include some fields from model.
        @return: (dict) with mixed values
        """
        if not self.is_payload(dict):
            raise TypeError(f"as_dict() can only be called when the payload data is a dictionary")

        if select_fields is not None:
            r = {}
            for k in select_fields:
                if isinstance(self._attr[k], self.__class__):
                    v = self._attr[k].as_dict()
                else:
                    v = self._attr[k]
                r[k] = v
            return r
        else:
            return {
                k: v.as_native() if isinstance(v, self.__class__) else v
                for k, v in self._attr.items()
            }

    def as_native(self):
        """
        @return: (mixed)
            representation of the payload (as children elements) comprised of
            native python data types.
        """
        if self.payload_undefined:
            return None

        if self.is_payload(dict):
            return self.as_dict()

        if self.is_payload(list):
            r = []
            for item in self._attr:
                value = item.as_native() if isinstance(item, self.__class__) else item
                r.append(value)
            return r

        if self.is_payload(set):
            r = set()
            for item in self._attr:
                value = item.as_native() if isinstance(item, self.__class__) else item
                r.add(value)
            return r

        raise TypeError("Unsupported type")

    def as_json(self, *args, **kwargs):
        """
        @see :method:`as_dict` for params.
        @returns (str) JSON representation
        """
        return json.dumps(self.as_native(*args, **kwargs), default=str)

    def __getattr__(self, attr):
        if attr not in self._attr:
            raise AttributeError(
                "{} instance has no attribute '{}'".format(self.__class__.__name__, attr)
            )
        if isinstance(self._attr[attr], list):

            def list_recurse(item):
                r = []
                for s in item:
                    if isinstance(s, dict):
                        r.append(self.__class__(s))
                    elif isinstance(s, list):
                        r.append(list_recurse(s))
                    else:
                        r.append(s)
                return r

            return list_recurse(self._attr[attr])

        elif isinstance(self._attr[attr], dict):
            return self.__class__(self._attr[attr])
        else:
            return self._attr[attr]

    def __setattr__(self, attr, val):
        super(Pinnate, self).__setattr__(attr, val)
        if attr != "_attr":
            if self.payload_undefined:
                self._attr = {}

            self._attr[attr] = val

    def __getitem__(self, key):
        return self._attr[key]

    def __setitem__(self, key, value):
        if self.payload_undefined:
            # if key is an integer the datatype *could* also be list
            self._attr = {}

        self._attr[key] = value

    def get(self, key, default=None):
        return self._attr.get(key, default)

    def load(self, data):
        """
        :param data: dict, list, set or json string
        :param merge: bool see :method:`update` if False or :method:`merge` when True.
        """

        if isinstance(data, str):
            data = json.loads(data)

        self.update(data)

    def update(self, data):
        """
        Extend the Pinnate with further payload values.

        If a setting with an existing key is supplied, then the previous value is overwritten.

        If the payload is a -
         - list - it will be extended
         - set - added to
         - dict - merged

        :param data: dictionary or list or set or json string
        """

        if not isinstance(data, (dict, list, set)):
            raise TypeError("Unsupported type")

        if self.payload_undefined:

            if isinstance(data, dict):
                self._attr = {}
            elif isinstance(data, set):
                self._attr = set()
            elif isinstance(data, list):
                self._attr = []

        if not self.is_payload(type(data)):
            p_type = str(type(self._attr))
            d_type = str(type(data))
            msg = (
                f"The type of the update data '{d_type}' doesn't match current payload's "
                f"type: '{p_type}'"
            )
            raise TypeError(msg)

        if self.is_payload(dict):
            for k, v in data.items():
                if isinstance(v, dict):
                    self._attr[k] = Pinnate(v)
                else:
                    self._attr[k] = v

        elif self.is_payload(list):

            for v in data:
                if isinstance(v, dict):
                    self._attr.append(Pinnate(v))
                else:
                    self._attr.append(v)

        elif self.is_payload(set):

            for v in data:
                if isinstance(v, dict):
                    self._attr.add(Pinnate(v))
                else:
                    self._attr.add(v)

    def append(self, item):
        """
        Can be used to add item when the payload is a list.
        """
        self.update([item])

    def add(self, item):
        """
        Can be used to add item when the payload is a list.
        """
        self.update(set([item]))



from datetime import datetime
import unittest

from ayeaye.pinnate import Pinnate


class TestPinnate(unittest.TestCase):
    def test_attrib_and_dict(self):
        a = Pinnate({"my_string": "abcdef"})
        self.assertEqual(a.my_string, "abcdef")
        self.assertEqual(a["my_string"], "abcdef")
        self.assertEqual(a.as_dict(), {"my_string": "abcdef"})

    def test_recurse(self):
        d = {"my_things": [1, 2, {"three": 3}]}
        a = Pinnate(d)
        p = a.my_things
        self.assertTrue(p[0] == 1 and p[1] == 2 and isinstance(p[2], Pinnate))
        self.assertEqual(p[2].three, 3)

    def test_variable_method_name(self):
        """
        check it's possible to give a variable the same name as an existing method.
        """
        d = {"as_dict": 1}
        p = Pinnate(d)
        self.assertEqual("{'as_dict': 1}", str(p.as_dict()))

    def test_as_json(self):
        d = {
            "number": 1,
            "string": "hello",
            "date": datetime.strptime("2020-01-15 10:34:12", "%Y-%m-%d %H:%M:%S"),
            "recurse_list": [{"abc": "def"}],
            "recurse_dict": {"ghi": {"jkl": "mno"}},
        }
        p = Pinnate(d)
        as_json = str(p.as_json())
        expected = (
            '{"number": 1, "string": "hello", "date": "2020-01-15 10:34:12", '
            '"recurse_list": [{"abc": "def"}], "recurse_dict": {"ghi": {"jkl": "mno"}}}'
        )
        self.assertEqual(expected, as_json)

    def test_recursive_lists(self):
        "bug found with list inside a list"

        d = {
            "name": "Stall Street",
            "type": "Line",
            "payload": [[-2.3597675, 51.3797509], [-2.359739, 51.3797783]],
            "tags": {
                "name": "Stall Street",
                "oneway": "yes",
                "highway": "unclassified",
                "motor_vehicle:conditional": "no @ (10:00-18:00)",
            },
        }
        p = Pinnate(d)
        # accessing the item triggered the bug
        self.assertEqual(-2.3597675, p.payload[0][0])

        # deeper recurse
        d = {"a": [[[{"b": "hello"}]]]}
        p = Pinnate(d)
        self.assertEqual("hello", p.a[0][0][0].b)

    def test_top_level_list(self):
        "test you can make a Pinnate from a top level list"

        l = ["a", "b", "c"]

        p = Pinnate(l)

        self.assertEqual(p[0], "a")
        self.assertEqual(p[1], "b")
        self.assertEqual(p[2], "c")

    def test_top_level_list_of_dict(self):
        "test you can make a Pinnate from a top level list of dictionaries"

        l = [{"first_name": "mary"}, {"last_name": "poppins"}]

        p = Pinnate(l)

        self.assertEqual(p[0].first_name, "mary")
        self.assertEqual(p[1].last_name, "poppins")

    def test_update_dict(self):
        "test you can update an existing dict into a Pinnate made from a dict"

        first_name = {"first_name": "mary"}
        p = Pinnate(first_name)

        last_name = {"last_name": "poppins"}
        p.update(last_name)

        self.assertEqual(p.first_name, "mary")
        self.assertEqual(p.last_name, "poppins")

    def test_update_list(self):
        "test you can update a list into a Pinnate made from a list"

        l_one = [1, 2]
        p = Pinnate(l_one)

        l_two = [2, 3]
        p.update(l_two)

        self.assertEqual(p[0], 1)
        self.assertEqual(p[1], 2)
        self.assertEqual(p[2], 2)
        self.assertEqual(p[3], 3)

    def test_top_level_iterator(self):
        "a list at the top level makes the Pinnate behave like a normal list"

        p = Pinnate(["a", "b", "c"])
        joined_values = "".join(p)

        self.assertEqual("abc", joined_values)
        self.assertFalse(isinstance(p, list), "It's a Pinnate, not a list")

        # same behaviour at 2nd level
        px = Pinnate({"xx": ["a", "b", "c"]})
        joined_values = "".join(px.xx)
        self.assertEqual("abc", joined_values)
        self.assertIsInstance(px.xx, list, "Second level really is a list")

    def test_late_construction(self):
        "Pinnate becomes a list or dict or set but not during construction"
        # list
        p = Pinnate()
        p.append("x")
        self.assertEqual("x", p[0])

        # set
        p = Pinnate()
        p.add("x")
        self.assertIn("x", p)

        # dict
        p = Pinnate()
        p.x = "y"
        self.assertEqual(p["x"], "y")

        # dict with set item
        p = Pinnate()
        p["x"] = "y"
        self.assertEqual(p["x"], "y")

        # nothing there so 'in' must be false
        p = Pinnate()
        self.assertFalse("x" in p)

    def test_set_item_list(self):
        "positional access to a list. .__setitem__ also used with dictionaries"
        p = Pinnate(["a", "b", "c"])
        self.assertEqual("a", p[0])

        p[0] = "x"
        self.assertEqual("x", p[0])


if __name__ == "__main__":
    unittest.main()



from ayeaye.connectors.engine_type_modifiers.abstract_modifier import AbstractModifier


class SmartOpenModifier(AbstractModifier):
    """
    Use a sub-set of smart-open (https://pypi.org/project/smart-open/) to-
    - open s3 blobs
    - transparently compress and de-compress
    - expand file patterns using wild cards

    for all sub-classes of :class:`FilesystemConnector`.
    """

    my_thing = "this is smart open"


# >>> from smart_open import open, register_compressor
# >>> with open('smart_open/tests/test_data/1984.txt.gz', 'rb', compression='disable') as fin:
# ...     print(fin.read(32))
#
# >>> from smart_open import open, register_compressor
# >>> with open('smart_open/tests/test_data/1984.txt.gzip', compression='.gz') as fin:
# ...     print(fin.read(32))
# It was a bright cold day in Apri



import copy
import json
import re
import warnings


class ConnectorResolver:
    """
    Lazy substitution of template parameters.

    Normal usage of this module is for a single global instance to be available to all instances of
    all the subclasses of :class:`DataConnector`. This is provided as-

    >>> from ayeaye.connect_resolve import connector_resolver

    This instance is given a mapping (i.e. key value pairs in a dictionary) or it is given a
    `resolver_callable` that will be called only when needed. This call is most likely to be by a
    subclasses of DataConnector (i.e. connections to datasets) when they access the `engine_url`
    attribute.

    Example using named attributes-

    >>> import ayeaye
    >>> from ayeaye.connectors.csv_connector import CsvConnector
    >>> ayeaye.connect_resolve.connector_resolver.add(build_id="20210616A")
    >>> x = CsvConnector(engine_url="csv:///data/build_{build_id}/product_codes.csv")
    >>> x.engine_url
    'csv:///data/build_20210616A/product_codes.csv'
    >>>

    Example using a callable to resolve a secret:

    def my_env_resolver(unresolved_engine_url):
        '''finds params starting like '{env_xxx}' and replaces it with the content of the
        environmental variable 'xxx'
        '''
        # .... code not shown

    >>> import ayeaye
    >>> from ayeaye.connectors.sqlalchemy_database import SqlAlchemyDatabaseConnector
    >>> ayeaye.connect_resolve.connector_resolver.add_secret(my_env_resolver)
    >>> x = SqlAlchemyDatabaseConnector(engine_url="mysql://root:{env_secret_password}@localhost/my_database")
    >>> x.engine_url
    'mysql://root:p4ssw0rd@localhost/my_database'
    >>>

    """

    def __init__(self):
        self._clear_state()

    def _clear_state(self):
        self.unnamed_callables = []
        self._attr = {}

    def brutal_reset(self):
        """
        Discard any callables or attributes that have been set. This method is really for unit
        tests to stop a failed test effecting the 'clear' state for a subsequent test. Be careful
        not to call this other times.
        """
        self._clear_state()

    def __getattr__(self, attr):
        if attr not in self._attr:
            return DeferredResolution(self, attr)
        return self._attr[attr]

    def needs_resolution(self, engine_url):
        return engine_url is None or (isinstance(engine_url, str) and "{" in engine_url)

    def resolve_engine_url(self, unresolved_engine_url):
        """
        old name, use :method:`resolve` instead
        """
        warnings.warn(
            "'resolve_engine_url' is deprecated use 'resolve' instead", DeprecationWarning
        )
        return self.resolve(unresolved_engine_url)

    def resolve(self, unresolved):
        """
        Fully resolve template variables to an string that contains no further parameters or raise
        an exception if not possible to resolve all.

        @param unresolved: (str) with templated variables. e.g. "csv://{my_variable}/abc.csv"

        @return: (str) engine_url
        """
        resolving = unresolved

        if not self.needs_resolution(resolving):
            # not a lot to do!
            return resolving

        for r_callable in self.unnamed_callables:
            resolving = r_callable(resolving)
            if not self.needs_resolution(resolving):
                return resolving

        for k, v in self._attr.items():
            template_var = f"{{{k}}}"
            if template_var in resolving:
                resolving = resolving.replace(template_var, v)
                if not self.needs_resolution(resolving):
                    return resolving

        # warning - don't put the partially resolved engine url into stack trace as it might
        # contain secretes.
        missing_vars = ",".join(re.findall("{.+?}", resolving))
        msg = (
            f"Couldn't fully resolve engine URL. Unresolved: {unresolved}. "
            f"Missing template variables are: {missing_vars}"
        )
        raise ValueError(msg)

    def add(self, *args, **kwargs):
        """
        @param *args: a callable that will have single argument (engine_url (str)) passed
            to it and must return (str) with anything that can be resolved having been resolved.
            There will be a chain of these resolvers that will be used in turn until there are no
            more parameters that need resolution.
        @param **kwargs: add named attributes. The key is the templated variable e.g. "..{key}..."
            and the value is a variable to substitute into place.
            For an example, see docstring for class.
        """
        for resolver_callable in args:
            assert callable(resolver_callable)
            self.unnamed_callables.append(resolver_callable)

        for attribute_name, attribute_value in kwargs.items():
            if attribute_name in self._attr:
                raise ValueError(f"Attempted to set existing attribute: {attribute_name}")

            if not isinstance(attribute_name, (int, str)):
                raise ValueError(
                    f"templated variable '{attribute_name}' needs to be string or int."
                )

            self._attr[attribute_name] = attribute_value

    def add_secret(self, *args, **kwargs):
        """
        When an :class:`ayeaye.Model` is locked secrets shouldn't be included in the locking data.
        This method is analogous to :method:`add` but keeps a reference to exclude these items
        from the :method:`capture_context`.

        @see :method:`add` for arguments.
        """
        raise NotImplementedError("TODO")

    def capture_context(self):
        """
        Return a JSON safe dictionary of context variables.

        The returned dictionary is expected to be serialised as part of :method:`ayeaye.Model.lock`.

        @return (dict)
        """
        json_safe_types = (str, int, float, bool)

        # TODO: secrets - these shouldn't be returned

        # work in progress. Can only handle static mappings and not callables.
        if len(self.unnamed_callables) > 0:
            # alt approach is for ayeaye.Model to implement load_locking and unload_locking methods
            # and for the mapper dictionary to support callables and overlay of locking info
            raise NotImplementedError("Can't serialise callables - alternative approach needed.")

        for attribute_name, attribute_value in self._attr.items():
            # attrib names are checked in :method:`add`
            if isinstance(attribute_value, json_safe_types):
                continue

            try:
                # sure fire way to check if it can be serialised into JSON
                _ = json.dumps(attribute_value)
            except TypeError:
                raise ValueError(f"Non-JSON serialisable data type found in '{attribute_name}'")

        # copy to make a snapshot as context manager will change _attr
        return {"mapper": copy.copy(self._attr)}

    def context(self, *args, **kwargs):
        """
        Use a resolver_callable just for the duration of a with statement. e.g.

        c = CsvConnector(engine_url="csv://my_path/data_{data_version}.csv")
        with connector_resolver.context(m_resolver):
            assert 'csv://my_path/data_1234.csv' == c.engine_url

        @see :method:`TestConnectors.test_resolve_engine_url` for an example.

        @see :method:`add` for args and kwargs

        Warning - not yet thread-safe as the global state is altered for the duration of the
        context manager. Can be fixed with thread local variables.
        """

        class _LocalResolverContext:
            """
            Keep track of a temporary resolver.

            Can be used with a 'with statement' or by calling :method:`start` and :method:`finish`.
            @see :method:`TestResolve.test_without_with_statement` for an example.
            """

            def __init__(self, parent):
                self.context_args = []
                self.context_kwargs = {}
                self.args_count = None
                self.named_attr = None
                self.parent = parent

            def add(self, *a_args, **a_kwargs):
                """
                @see :method:`ConnectorResolver.add`
                """
                if a_args:
                    self.context_args.extend(a_args)

                if a_kwargs:
                    self.context_kwargs.update(a_kwargs)

            def start(self):
                """
                Add a new resolver callable to the temporary context.
                """
                self.args_count = len(self.context_args)
                self.named_attr = self.context_kwargs
                self.parent.add(*self.context_args, **self.context_kwargs)

            def finish(self):
                """
                Clear this context away from the ConnectResolve instance.
                """
                # TODO check the right one is being cleared. This is currently making the assumption
                # that the last item is the right item.
                for _ in range(self.args_count):
                    del self.parent.unnamed_callables[-1]

                for attr_name in self.named_attr.keys():
                    del self.parent._attr[attr_name]

            def __enter__(self):
                self.start()

            def __exit__(self, exc_type, exc_val, exc_tb):
                self.finish()

        isolated_local_context = _LocalResolverContext(parent=self)
        isolated_local_context.add(*args, **kwargs)
        return isolated_local_context


class DeferredResolution:
    """
    Avoid the catch 22 of :class:`ConnectorResolver`'s named attribute being needed by a class variable
    but having not been added to the connector_resolver instance.

    Save the calling until :method:`Connect._prepare_connection calls` :method:`evaluate`.
    """

    def __init__(self, calling_instance, requested_attrib):
        self.calling_instance = calling_instance
        self.requested_attrib = requested_attrib

        # for now this isn't a general pattern, just one level down is a method which is passed
        # args and kwargs or it's a plain variable.
        self.second_level_attrib_name = None
        self.method_args = None
        self.method_kwargs = None

    def __getattr__(self, attrib_name):
        self.second_level_attrib_name = attrib_name

        def callable_might_be_needed(*args, **kwargs):
            self.method_args = args
            self.method_kwargs = kwargs
            return self

        return callable_might_be_needed

    def __call__(self):
        original_attrib = getattr(self.calling_instance, self.requested_attrib)
        if self.method_args or self.method_kwargs:
            # attrib was a method
            target_method = getattr(original_attrib, self.second_level_attrib_name)
            r = target_method(*self.method_args, **self.method_kwargs)
            return r
        else:
            # just return the attrib
            return original_attrib


# global provider of context
connector_resolver = ConnectorResolver()



from collections import defaultdict
import copy
from dataclasses import dataclass
from inspect import isclass
import itertools

from ayeaye.connectors.base import AccessMode
from ayeaye.connectors.multi_connector import MultiConnector
from ayeaye.model import Model
from ayeaye.pinnate import Pinnate


class ModelCollection:
    def __init__(self, models):
        """
        Build graphs from a group of subclasses of :class:`ayeaye.Models`

        @param models (class, list, set or callable)
            All of which are :class:`ayeaye.Models`. They aren't instances of the class but the
            class itself. A *list* is an explicit ordering that the models should be run in. Models
            in *set* will be examined to see dataset dependencies and these will be used to
            determine the model run order.
        """
        super().__init__()

        invalid_construction_msg = (
            "models must be a class, list, set or callable. All of which "
            "result in one or more :class:`ayeaye.Models` classes (not "
            "instances)."
        )

        # validate and prepare
        if isclass(models) and issubclass(models, Model):
            models = [
                models,
            ]

        if callable(models):
            # TODO
            raise NotImplementedError("TODO")

        elif isinstance(models, (list, set)):
            if not all([isclass(m) and issubclass(m, Model) for m in models]):
                raise ValueError(invalid_construction_msg)
            self.models = models

        else:
            raise ValueError(invalid_construction_msg)

    def __len__(self):
        return len(self.models)

    def __iter__(self):
        """
        yield model classes. No ordering when models are a set but does honour order when models
        is a list
        """
        yield from self.models

    def _base_graph(self):
        """Find all the datasets in all the models and classify datasets as 'sources' (READ access)
        and 'targets' (WRITE access) or READWRITE for both.

        @return: (targets, sources, nodes) (set, set, dictionary)

            node - 'datasets' are a subclasses of :class:`Connect` or :class:`DataConnector`

            all_targets - one or more models writes to these
            all_sources - one or more models read from these
            nodes - key is model class, value is Pinnate with .model_cls, .model_name, .targets, .sources
        """
        all_targets = set()
        all_sources = set()
        nodes = {}
        for model_cls in self.models:
            # TODO find ModelConnectors and recurse into those
            model_name = model_cls.__name__
            if model_name in nodes:
                raise ValueError(f"Duplicate node found: {model_name}")

            node = Pinnate({"model_cls": model_cls, "model_name": model_name, "targets": set(), "sources": set()})

            # as instantiated model
            # model_obj = model_cls()

            for class_attrib_label, ds_connector in model_cls.connects().items():
                # for class_attrib_label, ds_connector in model_obj.datasets().items():

                # if isinstance(ds_connector, Connect):
                #
                #     if isinstance(ds_connector, list):
                #         # experiment with Connector prior to instantiation as MultiConnector
                #         m_connectors = [Connect()]

                if isinstance(ds_connector, MultiConnector):
                    m_connectors = ds_connector.data
                else:
                    m_connectors = [ds_connector]

                for dataset_connector in m_connectors:
                    dataset_container = ModelDataset(
                        model_attrib_label=class_attrib_label, connector=dataset_connector
                    )

                    if dataset_connector.access in [AccessMode.READ, AccessMode.READWRITE]:
                        node.sources.add(dataset_container)
                        all_sources.add(dataset_container)

                    if dataset_connector.access in [AccessMode.WRITE, AccessMode.WRITE]:
                        node.targets.add(dataset_container)
                        all_targets.add(dataset_container)

            nodes[model_cls] = node

        return all_targets, all_sources, nodes

    def _resolve_run_order(self):
        """
        Use the dataset connections in each model to determine the dependencies between models and
        therefore the order to run them in.

        @return: (Pinnate) with attributes:
                    leaf_sources (set of datasets) - read but not written to
                    leaf_targets (set of datasets) - written to but not read - end goal of model
                    run_order (list of sets of nodes) - All models in a set can be run in parallel.
                            Each set must be complete before the next set is run.

                    'datasets' as subclasses of :class:`Connect`
                    `nodes` are type :class:`Pinnate` with attributes model_cls, model_name,
                        targets and sources
        """
        # This algorithm is a bit overly simplistic and sub-optimal. The return is a list of sets
        # with all models in set needing to be run. But the set is actually the models that were
        # ready to run when the prior models are complete. So there could be subsequent models that
        # are only waiting on a subset of each set.
        all_targets, all_sources, nodes = self._base_graph()

        leaf_sources = all_sources - all_targets
        leaf_targets = all_targets - all_sources

        completed = copy.copy(leaf_sources)
        run_order = []
        while len(nodes) > 0:
            loop_ready = set()
            for node in nodes.values():
                if node.sources.issubset(completed):
                    loop_ready.add(node)

            if len(loop_ready) == 0:
                msg = "The set of models can't be built into a single acyclic graph."
                raise ValueError(msg)

            # nodes in loop_ready have been visited add their edges to the completed map
            run_order.append(loop_ready)
            for ready_node in loop_ready:
                node = nodes[ready_node.model_cls]
                completed.update(node.targets)
                del nodes[ready_node.model_cls]

        p = Pinnate({"leaf_sources": leaf_sources, "leaf_targets": leaf_targets, "run_order": run_order})
        return p

    def run_order(self):
        """
        @return: (list) items in the list are either-
                    (i) sets of subclasses of :class:`ayeaye.Model`
                    OR
                    (ii) lists of sets as per (i)

        All models in a set can be run in parallel. Each set must be complete before the next set
        in the list is run.
        """
        if isinstance(self.models, list):
            # models in 'list' mode are wrapped as single item sets because all items in a set must
            # complete before next item in a list is run.
            return [{m} for m in self]

        elif isinstance(self.models, set):
            # _resolve_run_order returns leaf nodes and the run order is built using Pinnate
            # instances as nodes. Remove all this with to return a simple run order.
            resolved_order = self._resolve_run_order()

            def simplify_nodes(n):
                if isinstance(n, Pinnate):
                    return n.model_cls
                if isinstance(n, (set, list)):
                    rx = [simplify_nodes(nx) for nx in n]
                    if isinstance(n, set):
                        return set(rx)
                    return rx

            return [simplify_nodes(r) for r in resolved_order.run_order]

        else:
            raise ValueError("Unknown models container.")

    def dataset_provenance(self):
        """
        self.models is a set of :class:`ayeaye.Model`s. These might all be interconnected or they
        may form multiple graphs.

        @return: set (graphs) of sets (edges)
            graphs are a set of edges
            edges are :class:`ModelGraphEdge` objects.
        """
        all_targets, all_sources, nodes = self._base_graph()
        leaf_sources = all_sources - all_targets
        leaf_targets = all_targets - all_sources

        # lookup to sources for each dataset
        dataset_sources = defaultdict(list)
        for node in nodes.values():
            for dataset_container in node.sources:
                dataset_sources[dataset_container].append(node.model_cls)

        def dataset_source(target_dataset_container):
            "generator yielding model classes"
            for source_model_cls in dataset_sources[target_dataset_container]:

                for source_dataset_container in nodes[source_model_cls].sources:
                    if source_dataset_container == target_dataset_container:
                        yield source_model_cls, source_dataset_container

        edge_set = set()
        for node in nodes.values():

            # print(node.model_cls.__name__, len(node.targets))

            for dataset_container_a in node.sources:

                if dataset_container_a in leaf_sources:
                    dataset_label = dataset_container_a.model_attrib_label
                    mge = ModelGraphEdge(model_a=None, model_b=node.model_cls, dataset_label=dataset_label)
                    edge_set.add(mge)

            for dataset_container_a in node.targets:

                if dataset_container_a in leaf_targets:
                    dataset_label = dataset_container_a.model_attrib_label
                    mge = ModelGraphEdge(model_a=node.model_cls, model_b=None, dataset_label=dataset_label)
                    edge_set.add(mge)

                for model_b, dataset_container_b in dataset_source(dataset_container_a):

                    # print(node.model_cls, dataset_container_a, model_b, dataset_container_b)

                    dataset_label = dataset_container_a.model_attrib_label
                    # models might each use different attrib names
                    if dataset_container_b is not None and dataset_label != dataset_container_b.model_attrib_label:
                        dataset_label += " / " + dataset_container_b.model_attrib_label

                    mge = ModelGraphEdge(model_a=node.model_cls, model_b=model_b, dataset_label=dataset_label)
                    edge_set.add(mge)

        # TODO - are sub graphs needed?
        graph_set = list()
        graph_set.append(list(edge_set))
        return graph_set


@dataclass
class ModelDataset:
    """
    Container for attributes associated with a dataset that belongs to a model.

    A dataset could be a :class:`ayeaye.Connect` or it could have been resolved to a subclass of
    :class:`DataConnector`. Hold these (todo) and other associated info about the dataset.
    """

    model_attrib_label: str  # name of class variable in parent model
    connector: object  #  ayeaye.Connect

    def __hash__(self):
        """
        super critical to graphs being able to build is equating two datasets are the same thing
        without the name (i.e. model_attrib_label) mattering.
        """
        return self.connector.__hash__()

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.connector == other.connector
        return False


@dataclass
class ModelGraphEdge:
    """
    A representation of the :class:`ModelGraph` is a list of edges.

    Two :class:`ayeaye.Model`s and a 'dataset' (:class:`Connect` or `DataConnector`) is an edge.
    This was an arbitrary decision, it could have been two datasets and a model.
    """

    model_a: Model
    model_b: Model
    dataset_label: str

    def __hash__(self):
        return hash((self.model_a, self.model_b, self.dataset_label))


class VisualiseModels:
    "Experiment to visualise run order and data provenance for a :class:`ModelCollection` instance"

    def __init__(self, model_collection):
        """
        @param model_collection - instance of :class:`ModelCollection`
        """
        self.model_collection = model_collection

    def mermaid_data_provenance(self):
        """
        @return (str)
            mermaid format (see https://github.com/mermaid-js/mermaid#readme) to visualise model's
            data dependencies.
        """

        def _leaf_label():
            "return name (str) for a leaf node"
            r = 0
            while True:
                yield f"leaf_{r}([ ])"
                r += 1

        # no idea if leaves are the same dataset or not
        leaf_label = _leaf_label()

        graphset = self.model_collection.dataset_provenance()
        if len(graphset) == 0:
            return ""

        # if len(graphset) > 1:
        #     raise NotImplementedError("Multiple graphs within one collection of models is not yet implemented!")

        out = ["graph LR"]
        for graph in graphset:
            # graphs don't need to be separate for Mermaid
            for edge in graph:

                model_a = edge.model_a.__name__ if edge.model_a is not None else next(leaf_label)
                model_b = edge.model_b.__name__ if edge.model_b is not None else next(leaf_label)

                edge_fmt = f"{model_a}-->|{edge.dataset_label}| {model_b}"
                out.append(edge_fmt)

        return "\n".join(out)



"""
This link between the runtime 'execution' environment and a worker process running a model.

Created on 31 Jan 2022

@author: si
"""


class RuntimeKnowledge:
    """
    This link between the runtime 'execution' environment and a worker process running a model.

    This module will develop to make it possible for a model to introspect and communicate with
    it's environment.

    Available attributes-

    * worker_id (int) - unique integer assigned in ascending order to workers as they start
    * total_workers (int) - Number of workers created in worker group. Is `None` if variable number
                    of workers are being used.
    """

    def __init__(self):
        self.worker_id = None
        self.total_workers = None



from io import StringIO
import json
import os
import shutil
import tempfile
import unittest

import ayeaye

from examples.favourite_colours import FavouriteColours
from examples.poisonous_animals import PoisonousAnimals
from examples.manifest_mapper import AustralianAnimals

EXAMPLE_MODELS_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'examples')
EXAMPLE_DATA = os.path.normpath(os.path.join(EXAMPLE_MODELS_DIR, 'data'))


class TestExamples(unittest.TestCase):

    def setUp(self):
        self._working_directory = None

        # the example models use relative paths to the example data
        self.current_working_directory = os.getcwd()
        os.chdir(EXAMPLE_MODELS_DIR)

    def tearDown(self):
        if self._working_directory and os.path.isdir(self._working_directory):
            shutil.rmtree(self._working_directory)

        os.chdir(self.current_working_directory)

    def working_directory(self):
        self._working_directory = tempfile.mkdtemp()
        return self._working_directory

    def test_poisonous_animals_fetch_data(self):
        """
        This model reads a local .flowerpot file. Run it and check the logged output (as it doesn't
        have any output datasets).
        """
        # relative path is used in connection
        external_log = StringIO()

        m = PoisonousAnimals()
        m.set_logger(external_log)
        m.log_to_stdout = False
        m.go()

        external_log.seek(0)
        all_the_logs = external_log.read()
        expected = 'In Australia you could find Blue ringed octopus,Box jellyfish,Eastern brown snake'
        self.assertIn(expected, all_the_logs)

    def test_favourite_colours_build_sample_data(self):
        """
        Using .build() and not .go() so pre and post checks aren't tested.
        """
        output_file = "{}/favourite_colours_summary.json".format(self.working_directory())
        m = FavouriteColours()
        m.log_to_stdout = False

        # give the connector a new output file
        m.favourites_summary.update(engine_url=f"json://{output_file};indent=4")
        m.build()
        output_encoding = m.favourites_summary.encoding
        m.close_datasets()

        with open(output_file, 'r', encoding=output_encoding) as f:
            output_data = json.load(f)

        # check Blue which has dates 2020-01-01 - 2020-02-15
        self.assertEqual(31, output_data["Blue"]["January"])
        self.assertEqual(14, output_data["Blue"]["February"])

    def test_favourite_colours_pre_post_checks(self):
        """
        Check .pre_build_check() and .post_build_check()
        """
        output_file = "{}/favourite_colours_summary.json".format(self.working_directory())
        m = FavouriteColours()
        m.log_to_stdout = False

        # give the connector a new output file
        m.favourites_summary.update(engine_url=f"json://{output_file}")
        self.assertTrue(m.go(), "Pre, post and build should work for favourite_colours.csv")

        # now run it with bad data.
        m = FavouriteColours()
        m.log_to_stdout = False
        external_log = StringIO()
        m.set_logger(external_log)

        m.favourite_colours = ayeaye.Connect(engine_url='csv://data/favourite_colours_bad_data.csv')
        m.favourites_summary.update(engine_url=f"json://{output_file}")

        # There are two issues.
        # (i) multiple years aren't supported, this should be caught by the pre_build_check
        self.assertFalse(m.pre_build_check(), "favourite_colours_bad_data.csv should fail.")
        m.close_datasets()  # reset file pointers
        external_log.seek(0)
        all_the_logs = external_log.read()
        expected = 'This model is only designed to work with data from a single year.'
        self.assertIn(expected, all_the_logs)

        # (ii) conservation of value - total number of days in the output should match the number of
        # days in the input. This fails because the algorithm in build assumes the data is good and
        # start date is before end date. The bad data file has one pair swapped.
        m.build()  # run it anyway
        m.close_datasets()  # reset file pointers
        self.assertFalse(m.post_build_check(), "favourite_colours_bad_data.csv should fail.")
        external_log.seek(0)
        all_the_logs = external_log.read()
        expected = "Total days in input doesn't match total days in output."
        self.assertIn(expected, all_the_logs)

    def test_australian_animal_mapper(self):

        c = {'input_path': EXAMPLE_DATA,
             'output_path': self.working_directory()
             }

        with ayeaye.connector_resolver.context(**c):
            m = AustralianAnimals()
            m.log_to_stdout = False
            m.go()

        # two files will be built, only one is checked
        poisonous_output = os.path.join(c['output_path'], 'australian_poisonous_animals.json')
        with open(poisonous_output) as f:
            # just as text, don't de-serialise the JSON
            poisonous = f.read()

        self.assertIn('Box jellyfish', poisonous)
        self.assertNotIn('Arizona Bark Scorpion', poisonous)



import unittest

import ayeaye
from ayeaye.runtime.multiprocess import ProcessPool


class ExamineResolverContext(ayeaye.PartitionedModel):
    def fake_subtask(self):

        try:
            ayeaye.connector_resolver.resolve("{build_environment_variable}")
            build_environment_variable_set = True
        except ValueError:
            build_environment_variable_set = False

        try:
            ayeaye.connector_resolver.resolve("{local_variable}")
            local_variable_set = True
        except ValueError:
            local_variable_set = False

        r = dict(
            build_environment_variable_set=build_environment_variable_set,
            local_variable_set=local_variable_set,
        )
        return r


class TestRuntimeMultiprocess(unittest.TestCase):
    """
    Test the local execution of :class:`ayeaye.Model`s with multiple processes.
    """

    def test_resolver_context_not_inherited(self):
        """
        The resolver context shouldn't be copied to worker processes from the parent. It should be
        explicitly passed in :method:`ayeaye.runtime.multiprocess.ProcessPool.run_model`.

        In practice they will be the same thing. This behaviour is checked as there was an issue
        with OSX fork() and Linux fork() having differences in memory from parent process? i.e.
        OSX didn't have the context and Linux did.
        """
        workers_count = 1

        with ayeaye.connector_resolver.context(build_environment_variable="is set"):

            msg = "Normal Aye-aye context resolver just sees vars in ayeaye.connector_resolver singleton"
            m = ExamineResolverContext()
            results = m.fake_subtask()
            self.assertTrue(results["build_environment_variable_set"], msg)
            self.assertFalse(results["local_variable_set"], msg)

            msg = "build_environment_variable should be discarded, local_variable should be present"
            proc_pool = ProcessPool(
                processes=workers_count,
                context_kwargs={
                    "mapper": {"local_variable": "is_set"},
                },
            )
            for method_name, method_kwargs, subtask_return_value in proc_pool.run_subtasks(
                model_cls=ExamineResolverContext, tasks=[("fake_subtask", None)], initialise=None
            ):
                # there will be only one sub-task return - this isn't tested
                self.assertEqual(method_name, "fake_subtask")
                self.assertEqual(method_kwargs, {})
                self.assertFalse(subtask_return_value["build_environment_variable_set"], msg)
                self.assertTrue(subtask_return_value["local_variable_set"], msg)



"""
Created on 17 Dec 2020

@author: si
"""
try:
    import ndjson
except ModuleNotFoundError:
    pass

from ayeaye.connectors.base import AccessMode, FileBasedConnector, FilesystemEnginePatternMixin
from ayeaye.pinnate import Pinnate


class NdjsonConnector(FileBasedConnector, FilesystemEnginePatternMixin):
    engine_type = "ndjson://"
    default_character_encoding = "utf-8-sig"

    def __init__(self, *args, **kwargs):
        """
        Connector to Newline Delimited JSON files (http://ndjson.org/).

        ndjson is one json document per line.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for NdjsonConnector
         None

        Connection information-
            engine_url format is
            ndjson://<filesystem absolute path>[;encoding=<character encoding>]
        e.g. ndjson:///data/my_project/the_data.json;encoding=latin-1
        """
        super().__init__(*args, **kwargs)

        self._reset()

        if self.access == AccessMode.READWRITE:
            raise NotImplementedError("READWRITE access not yet implemented")

    def _reset(self):
        FileBasedConnector._reset(self)
        self.reader = None
        self.writer = None
        self.approx_position = 0

    def connect(self):
        if self.reader is None and self.writer is None:
            if self.access == AccessMode.READ:
                FileBasedConnector.connect(self)
                self.reader = ndjson.reader(self._file_handle)

            elif self.access == AccessMode.WRITE:
                FileBasedConnector.connect(self)
                self.writer = ndjson.writer(self._file_handle)

            else:
                raise ValueError("Unknown access mode")

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        self.connect()

        for r in self.reader:
            # OSError: telling position disabled by next() call so this for now.
            # TODO: It's a waste of CPU to make it back into a string.
            self.approx_position += len(str(r))
            yield Pinnate(data=r)

    @property
    def data(self):
        raise NotImplementedError("TODO")

    def add(self, data):
        """
        Write record to ndjson file.
        @param data: (dict or Pinnate) - must be safe to serialise to JSON so no dates etc.
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        self.connect()

        if isinstance(data, dict):
            self.writer.writerow(data)
        elif isinstance(data, Pinnate):
            self.writer.writerow(data.as_dict())
        else:
            raise ValueError("data isn't an accepted type. Only (dict) or (Pinnate) are accepted.")



from ayeaye.connectors.base import DataConnector


class FakeDataConnector(DataConnector):
    """
    Fake connector used in unit tests.
    """

    engine_type = "fake://"
    optional_args = {
        "quantum_accelerator_module": None,
        "quantum_factory": None,
    }
    preserve_callables = ["quantum_factory"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def engine_params(self):
        """
        :returns: (dict) of parameters needed to connect to engine without secrets.
        """
        return {"engine_url": "fake://example.com/abc"}

    def connect(self):
        pass

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    @property
    def data(self):
        return [{"fake": "data"}]

    @property
    def schema(self):
        return None



"""
Created on 14 Jul 2022

@author: si
"""
import json
from time import time

try:
    import requests
except ModuleNotFoundError:
    pass


from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class RestfulConnector(DataConnector):
    engine_type = ["http://", "https://"]
    optional_args = {
        "headers": None,
        "keep_alive": True,  # http 1.1 mode is the default
    }

    def __init__(self, *args, **kwargs):
        """
        Connector to read and write to JSON RestFul APIs.

        See-
        * https://en.wikipedia.org/wiki/Representational_state_transfer
        * https://jsonapi.org/

        For args: @see :class:`connectors.base.DataConnector`

        additional args for RestfulConnector
            headers (dictionary) - key value pairs to pass in the http header. These are often used
                                for authentication keys.
            keep_alive (Boolean) - use sessions to persist cookies and reuse TCP sockets. This mode
                                also enables retry on fail to connect.

        Connection information-
            engine_url format is
            http[s]://<base_url>
            e.g. https://twitter.com
            all requests made by this connector are expected to start with the http[s]://base_url
            self.engine_url in the code is this URI
        """
        super().__init__(*args, **kwargs)

        # config flags - these could become optional args later. Hard coded for now
        # --------------------------
        # Request is retired for these HTTP status codes
        self.retry_on_status = [
            500,
            502,
            503,
            504,
        ]
        self.raise_exception_on_404 = False
        self.raise_exception_on_400s = True
        self.raise_exception_on_500 = True
        self.max_retries = 3
        # --------------------------

        # updated whilst running
        self.last_http_status = None
        self.last_headers = None
        self._requests = None  # initiated by :method:`connect`
        self.statistics = None  # OK to read this
        self.reset_stats()
        self.profiler = ProfileRequest(statistics=self.statistics)

    def reset_stats(self):
        """clear profiling requests made during lifetime of object
        - 'requests_total_time' is time spent waiting for web server (sec)
        """

        self.statistics = Pinnate(
            {
                "requests_count": 0,
                "requests_total_time": 0.0,
                "requests_slowest_time": None,
                "requests_slowest_url": None,
            }
        )

    def close_connection(self):
        if self._requests is not None:
            self._requests = None
        self.reset_stats()

    def connect(self):
        if self._requests is None:
            if self.keep_alive:
                self._requests = requests.Session()

                retry_config = requests.adapters.Retry(
                    total=self.max_retries,
                    backoff_factor=0.1,
                    status_forcelist=self.retry_on_status,
                )

                self._requests.mount(
                    "http://", requests.adapters.HTTPAdapter(max_retries=retry_config)
                )
                self._requests.mount(
                    "https://", requests.adapters.HTTPAdapter(max_retries=retry_config)
                )
            else:
                self._requests = requests

    def __len__(self):
        """
        Doesn't make sense for a RestFul API
        """
        raise NotImplementedError("Not available for API connector")

    def __getitem__(self, key):
        """
        Doesn't make sense for a RestFul API
        """
        raise NotImplementedError("Not available for API connector")

    def __iter__(self):
        """
        Doesn't make sense for a RestFul API
        """
        raise NotImplementedError("Not available for API connector")

    @property
    def schema(self):
        raise NotImplementedError("Not sure if possible for the API connector")

    @property
    def not_found(self):
        """
        @return: bool
            Return True if the last API request was for an entity that didn't exist.
        """
        return self.last_http_status == 404

    @property
    def response_headers(self):
        """
        @return: list of str
            http headers from server for last request.
        """
        return self.last_headers

    def qualify_url(self, url):
        """
        Ensure URL is fully formed so is ready to pass to requests library.

        @param url: str
            either url without self.engine_url or already a fully qualified url

        @return: str
            fully qualified url.
        """
        if url.startswith("http://") or url.startswith("https://"):
            msg = (
                "RestfulConnector needs all requests to start with the same URL base. Current"
                f" set to {self.engine_url}, this url is {url}/"
            )
            assert url.startswith(self.engine_url), msg
            return url
        else:
            return self.engine_url + url

    def get(self, url, params=None):
        """
        HTTP GET

        @param url: str
            either fully qualified or url suffix (e.g. 'site/' that needs self.engine_url)

        @param params: dictionary or :class:`Pinnate` of arguments to append to the URL.
                        e.g.
                        self.engine_url="https://api.mystuff.com/abc
                        c.get("/xyz", {'q':'search_term'})
                         becomes
                        https://api.mystuff.com/abc/xyz?q=search_term

        @return: :class:`Pinnate` object representing the JSON response
        """
        if self.access not in (AccessMode.READ, AccessMode.READWRITE):
            raise ValueError("Read attempted on dataset not opened in READ mode.")

        self.connect()

        if isinstance(params, Pinnate):
            _params = params.as_dict()
        else:
            _params = params

        url_ = self.qualify_url(url)
        with self.profiler(url_, _params):
            try:
                r = self._requests.get(url_, params=_params, headers=self.headers)
            except requests.ConnectionError as c:
                msg = f"Failed to GET from {url_}"
                raise RestfulConnectorConnectionError(msg, details=str(c))

        self._post_request_checks(r)
        serialised_request = Pinnate(r.json())
        return serialised_request

    def post(self, url, data):
        """
        HTTP POST

        @param url: str
            either fully qualified or url suffix (e.g. 'site/' that needs self.engine_url)

        @param data: :class:`Pinnate` object or something that serialises to JSON

        @return: :class:`Pinnate` object representing the JSON response or None if response
            doesn't contain valid JSON.
        """
        if self.access not in (AccessMode.WRITE, AccessMode.READWRITE):
            raise ValueError("Write attempted on dataset not opened in write mode.")

        self.connect()

        url_ = self.qualify_url(url)
        headers = {"Content-type": "application/json"}

        if isinstance(self.headers, dict):
            headers.update(self.headers)

        if isinstance(data, Pinnate):
            json_data = json.dumps(data.as_dict())
        else:
            json_data = json.dumps(data)

        with self.profiler(url_):
            try:
                r = self._requests.post(url_, data=json_data, headers=headers)
            except requests.ConnectionError as c:
                msg = f"Failed to POST to {url_}"
                raise RestfulConnectorConnectionError(msg, details=c.message)

        self._post_request_checks(r)

        # reply doc. is optional
        try:
            reply_doc = r.json()
        except requests.exceptions.JSONDecodeError:
            # r.text could contain something but probably empty
            return None

        serialised_request = Pinnate(reply_doc)
        return serialised_request

    def patch(self, url, data):
        """
        HTTP PATCH

        @param data: Pinnate object or something that serialises to JSON
        """
        if self.access not in (AccessMode.WRITE, AccessMode.READWRITE):
            raise ValueError("Write attempted on dataset not opened in write mode.")

        self.connect()

        url_ = self.qualify_url(url)
        headers = {"Content-type": "application/json"}

        if isinstance(self.headers, dict):
            headers.update(self.headers)

        if isinstance(data, Pinnate):
            json_data = json.dumps(data.as_dict())
        else:
            json_data = json.dumps(data)

        if isinstance(data, Pinnate):
            json_data = json.dumps(data.as_dict())
        else:
            json_data = json.dumps(data)

        with self.profiler(url_):
            try:
                r = self._requests.patch(url_, data=json_data, headers=headers)
            except requests.ConnectionError as c:
                msg = f"Failed to PATCH {url_}"
                raise RestfulConnectorConnectionError(msg, details=c.message)

        self._post_request_checks(r)

        # reply doc. is optional
        try:
            reply_doc = r.json()
        except requests.exceptions.JSONDecodeError:
            # r.text could contain something but probably empty
            return None

        serialised_request = Pinnate(reply_doc)
        return serialised_request

    def _post_request_checks(self, r):
        """
        Update flags etc. after each request
        """
        self.last_http_status = r.status_code
        self.last_headers = r.headers
        msg = f"Received http status: {r.status_code}"

        if self.last_http_status >= 400:
            try:
                as_json = r.json()
                detailed_message = as_json["message"]
            except:
                detailed_message = None

        if r.status_code == 500 and self.raise_exception_on_500:
            raise RestfulConnectorConnectionError(
                msg, details=detailed_message, last_http_code=r.status_code
            )

        if r.status_code == 404 and self.raise_exception_on_404:
            raise RestfulConnectorConnectionError(
                msg, details=detailed_message, last_http_code=r.status_code
            )

        if (
            r.status_code != 404
            and r.status_code >= 400
            and r.status_code < 500
            and self.raise_exception_on_400s
        ):
            raise RestfulConnectorConnectionError(
                msg, details=detailed_message, last_http_code=r.status_code
            )


class ProfileRequest:
    """
    Track time taken for API requests.

    Tied into RestfulConnector through statistics so is really a private class
    """

    def __init__(self, statistics):
        """
        @param statistics: Pinnate with attribs -  requests_count,
                requests_total_time, requests_slowest_time,
                requests_slowest_url
        """
        self.statistics = statistics
        self.url = None

    def __call__(self, url, params=None):
        """
        @param url: (str)
            used in statistics
        """
        self.url = url

        # TODO just dumping params to string. This isn't a correct URL encoded url
        if params:
            self.url += str(params)

        return self

    def __enter__(self):
        self.start = time()
        return self

    def __exit__(self, type_, value, traceback):
        elapsed = time() - self.start
        self.statistics.requests_count += 1
        self.statistics.requests_total_time += elapsed

        if (
            self.statistics.requests_slowest_time == None
            or elapsed > self.statistics.requests_slowest_time
        ):
            self.statistics.requests_slowest_time = elapsed
            self.statistics.requests_slowest_url = self.url


class RestfulConnectorConnectionError(Exception):
    def __init__(self, message, details=None, last_http_code=None):
        super().__init__(message)
        self.message = message
        self.details = details
        self.last_http_code = last_http_code

    def __str__(self):
        if self.details:
            return "%s [%s]" % (self.message, self.details)
        return self.message



"""
Common patterns/helpers for populating the resolver context.

@see ayeaye.connector_resolver

The resolver context is there to provide runtime variables to dataset connections (i.e. 
ayeaye.Connect(..)) within ayeaye.Models.
"""
import json
import os
import sys

import ayeaye


def manifest_build_context():
    """
    A manifest is a dataset document with variables needed to repeat a build. i.e. exact versions
    of files used, paths etc.

    This method makes all the variables in the manifest available within the resolver context. It
    also makes details of the manifest file available-
    * manifest_path - Filesystem path. e.g. /data/
    * manifest_file - absolute path to manifest file. e.g. /data/manifest_20220511.json

    For example of this working, see ayeaye/examples/manifest_build_context/

    TLDR;
    At the command line -
    $ python sensor_model.py manifest_file.json

    Warning: don't let the manifest file get too big as it will be loaded by every model. If you need big
    listings of files create secondary manifest files and refer to them in the primary manifest file.

    @return: ayeaye.connector_resolver context object
    """

    if len(sys.argv) != 2:
        msg = (
            "Manifest build context requires a single command line argument.\n"
            f"usage: python {sys.argv[0]} <manifest_file>\n"
            "e.g.\n"
            f"python {sys.argv[0]} /data/manifest_20220511.json"
        )
        sys.stderr.write(msg)
        sys.exit(1)

    manifest_file = os.path.abspath(sys.argv[1])
    manifest_path = os.path.dirname(manifest_file)

    with open(manifest_file, encoding="utf-8-sig") as f:
        manifest = json.load(f)

    context_values = {
        "manifest_path": manifest_path,
        "manifest_file": manifest_file,
    }

    # values in the manifest fil eoverwrite the manifest_* vars
    if not isinstance(manifest, dict):
        raise ValueError("The manifest file must contain a dictionary so key values can be used as context variables")
    context_values.update(manifest)

    context = ayeaye.connector_resolver.context(**context_values)

    return context



class AbstractModifier:
    """
    engine_urls can be prefixed with modifier that slightly changes how the underlying
    DataConnector works.

    For example-
        engine_url="gz+ndjson:///data/myfile.ndjson.gz"

    Would transparently decompress on the fly but is otherwise using the
    :class:`ayeaye.Connectors.NdjsonConnector` connector.

    Modifier classes are subclasses of this `AbstractModifier` class and their :method:`apply` is
    applied to the target `DataConnector` in the same way that python decorators work behind the
    scenes-

    e.g.

    >>> a_modifer = ExampleCloudStorageModifier()
    >>> my_connector = NdjsonConnector(**some_kwargs)
    >>> my_connector = a_modifer.apply(connector=my_connector)
    """

    my_thing = "hello"

    def apply(self, connector):
        """
        Modify a data connector to use this (i.e. `self`) modifier.

        WARNING - this method mutates the connector and returns the mutated connector.

        @param connector: instance of a subclass of :class:`DataConnector`
        @return: instance of a subclass of :class:`DataConnector`
        """
        raise NotImplementedError("Must be implemented by subclasses")



"""
Created on 15 Apr 2020

@author: si
"""
import json
import os

from ayeaye.connectors.base import AccessMode, FileBasedConnector, FilesystemEnginePatternMixin
from ayeaye.pinnate import Pinnate


class JsonConnector(FileBasedConnector, FilesystemEnginePatternMixin):
    engine_type = "json://"
    optional_engine_url_args = FileBasedConnector.optional_engine_url_args + ["indent"]
    default_character_encoding = "utf-8-sig"

    def __init__(self, *args, **kwargs):
        """
        Single JSON file loaded into memory and made available as a :class:`Pinnate` object.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for JsonConnector
         None

        Connection information-
            engine_url format is
            json://<filesystem absolute path>[;encoding=<character encoding>][;indent=<spaces when pretty printing write output>]
        e.g. json:///data/my_project/the_data.json;encoding=latin-1

        """
        self._reset()
        super().__init__(*args, **kwargs)

    @property
    def engine_params(self):
        if self._engine_params is None:
            self._engine_params = self._build_engine_params()

            for typed_param in ["indent"]:
                if typed_param in self._engine_params:
                    self._engine_params[typed_param] = int(self._engine_params[typed_param])

        return self._engine_params

    def _reset(self):
        FileBasedConnector._reset(self)
        self._doc = None

    def connect(self):
        """
        When in AccessMode.READ, read the contents of the file into a :class:`Pinnate` instance
        that is available as self.data.

        In AccessMode.WRITE mode connect() doesn't do anything because file handles aren't kept
        open by the JsonConnector. The write operation is in :method:`_data_write`.
        """
        if self._doc is None:
            FileBasedConnector.connect(self)

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        raise NotImplementedError("Not an iterative dataset. Use .data instead.")

    def _data_read(self):
        if self._doc is None:
            self.connect()
            as_native = json.load(self._file_handle)
            self._doc = Pinnate(as_native)

            if self.access == AccessMode.READWRITE:
                self._file_handle.seek(0)

        return self._doc

    def _data_write(self, new_data):
        """
        Set the contents of a JSON file. `new_data` can be an instance of :class:`Pinnate` or any
        python datatype that will serialise into JSON.

        Will raise TypeError if the data can't be serialised to JSON.

        @param new_data: (mixed, see description)
        """
        if self.access not in [AccessMode.WRITE, AccessMode.READWRITE]:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        self.connect()

        json_args = {}
        if "indent" in self.engine_params:
            json_args["indent"] = self.engine_params["indent"]

        if isinstance(new_data, Pinnate):
            as_json = json.dumps(new_data.as_dict(), **json_args)
        else:
            as_json = json.dumps(new_data, **json_args)

        # Data is written to beginning of file (it might be readwrite or already written to);
        # write to disk immediately (i.e. flush); @see :method:`connect`.
        self._file_handle.seek(0)
        self._file_handle.write(as_json)
        # truncate rest of the file as the previous contents might have been longer
        self._file_handle.truncate()
        self._file_handle.flush()

    data = property(fget=_data_read, fset=_data_write)

    @property
    def datasource_exists(self):
        """
        Returns:
            (bool) if the datasource referred to in self.engine_url exists.
        """
        return os.path.exists(self.engine_params.file_path)



from enum import Enum
import os.path

from ayeaye.connect_resolve import connector_resolver
from ayeaye.pinnate import Pinnate


class EngineUrlCase(Enum):
    # unprocessed first sight of engine_url - e.g. as passed to :class:`Connect`
    RAW = "raw"

    # everything needed to connect to dataset apart from secrets
    WITHOUT_SECRETS = "without_secrets"

    # has everything needed to connect to dataset
    FULLY_RESOLVED = "fully_resolved"


class EngineUrlStatus(Enum):
    OK = "ok"  # has been resolved
    NOT_AVAILABLE = "not_available"  # will never be available
    NOT_YET_AVAILABLE = "not_yet_available"  # deferred. waiting on something
    UNKNOWN = "unknown"


class Ignition:
    """
    Control engine_urls.
    - common functions like breaking the url into parts
    - resolving variables at runtime
    - translating path separators so engine_urls are independent of the operating system.

    'engine_urls' are the connection strings used by all subclasses of :class:`DataConnector`.

    This class is only used internally by subclasses of :class:`DataConnector`.
    """

    def __init__(self, engine_url):
        """
        @param engine_url: (str) an unresolved engine_url or one that doesn't need any resolution.
                An unresolved engine_url is one containing variables that need to be substituted
                before the engine_url can be used to connect to the dataset.
                e.g. "mysql://root:{env_secret_password}@localhost/my_database"
        """
        self._engine_url_state = {EngineUrlCase.RAW: engine_url}

    def engine_url_at_state(self, engine_url_case):
        """
        Get the engine_url in a specific case. e.g. raw, without secrets or fully resolved.

        @raise ValueError: from here or connector_resolver
        @param engine_url_case: (EngineUrlCase)
        @return: (EngineUrlStatus, str or None)
        """
        assert isinstance(engine_url_case, EngineUrlCase)

        if engine_url_case == EngineUrlCase.WITHOUT_SECRETS:
            raise NotImplementedError("TODO")

        if engine_url_case not in self._engine_url_state:
            raw_e_url = self._engine_url_state[EngineUrlCase.RAW]

            if raw_e_url is None:
                raise ValueError("engine_url passed as None")

            if connector_resolver.needs_resolution(raw_e_url):
                # local resolution is still needed
                resolved = connector_resolver.resolve(raw_e_url)
                self._engine_url_state[EngineUrlCase.FULLY_RESOLVED] = resolved
            else:
                # nothing to resolve so fully_resolved = raw
                self._engine_url_state[EngineUrlCase.FULLY_RESOLVED] = self._engine_url_state[
                    EngineUrlCase.RAW
                ]

        if engine_url_case in self._engine_url_state:
            return EngineUrlStatus.OK, self._engine_url_state[engine_url_case]

        return EngineUrlStatus.NOT_AVAILABLE, None

    @staticmethod
    def _decode_filesystem_engine_url(engine_url, required_args=None, optional_args=None):
        """
        For connectors that access files on a local filesystem break the URL into useful parts.

        URL format is
        engine_type://filesystem_path[;arg_0=argv_0[;...]]

        Raises value error if there is anything odd in the URL.

        @param engine_url: (str)
        @param required_args (list of str)
        @param optional_args (list of str)
        @return: (Pinnate) with .file_path and .engine_type
                                and optional and required args
        """
        if optional_args is None:
            optional_args = []

        if required_args is None:
            required_args = []

        all_args = optional_args + required_args

        # TODO cls.engine_type could be a list. It's normally a string.
        engine_type, remaining_url = engine_url.split("://", 1)
        path_plus = remaining_url.split(";")
        file_path = path_plus[0]
        d = {
            "file_path": file_path,
            "engine_type": engine_type,
        }
        if len(path_plus) > 1:
            for arg in path_plus[1:]:
                k, v = arg.split("=", 1)
                if k not in all_args:
                    raise ValueError(f"Unknown option: {k}")
                d[k] = v

        if os.path.sep != "/":
            # urrgh, Windoze
            d["file_path"] = d["file_path"].replace("/", os.path.sep)

        return Pinnate(d)



from collections import defaultdict
import json
import os
import shutil
import tempfile
import unittest

import ayeaye

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_CSV_PATH = os.path.join(PROJECT_TEST_PATH, "data", "deadly_creatures.csv")
EXAMPLE_TSV_PATH = os.path.join(PROJECT_TEST_PATH, "data", "monkeys.tsv")


class FindLongestAnimalName(ayeaye.PartitionedModel):
    """
    Find the longest common name in a collection of CSV/TSV files. Model suggests to executor how
    to break the task into parallel sub-tasks.
    """

    animals = ayeaye.Connect(
        engine_url=[
            f"csv://{EXAMPLE_CSV_PATH}",
            f"tsv://{EXAMPLE_TSV_PATH}",
        ]
    )
    animals_output = ayeaye.Connect(access=ayeaye.AccessMode.WRITE)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.common_names_max = []

    def build(self):
        pass

    def partition_plea(self):
        dataset_files_count = len(self.animals)
        return ayeaye.PartitionedModel.PartitionOption(
            minimum=1, maximum=dataset_files_count, optimal=dataset_files_count
        )

    def partition_slice(self, slice_count):

        target_method = "find_longest_name"  # this is the method subtasks should be running
        task_sets = defaultdict(list)
        for idx, dataset in enumerate(self.animals):
            task_id = idx % slice_count
            task_sets[task_id].append(dataset.engine_url)

        return [(target_method, {"engine_set": engine_set}) for engine_set in task_sets.values()]

    def partition_subtask_complete(self, subtask_method_name, subtask_kwargs, subtask_return_value):
        if subtask_method_name == "find_longest_name":
            self.common_names_max.append(subtask_return_value)

    def partition_complete(self):
        longest_animal_name = max(self.common_names_max, key=len)
        self.animals_output.data = longest_animal_name

    def find_longest_name(self, engine_set):
        """
        Find the longest common name in the self.animals[engine_url] dataset and return that
        string.
        """
        longest = ""
        for engine_url in engine_set:
            dataset = self.animals[engine_url]
            for row in dataset:
                if len(row.common_name) > len(longest):
                    longest = row.common_name

        return longest


class DistributedFakeWork(ayeaye.PartitionedModel):
    """
    Distribute a fake calculation and assemble the results.
    """

    # uses connector_resolver
    non_existant_data = ayeaye.Connect(engine_url="file://{hello_partitioned_context}")

    def build(self):
        pass

    def some_work(self, some_number):
        "add the worker id to a test number and append to fully resolved engine_url"
        some_data = self.non_existant_data.file_path + str(some_number)
        return some_data

    def partition_slice(self, _):
        target_method = "some_work"
        return [(target_method, {"some_number": x}) for x in range(10)]

    def partition_subtask_complete(self, subtask_method_name, subtask_kwargs, subtask_return_value):

        if not hasattr(self, "resultset"):
            self.resultset = []

        if subtask_method_name == "some_work":
            self.resultset.append(subtask_return_value)


class TestPartitionedModel(unittest.TestCase):
    def setUp(self):
        self._working_directory = None

    def tearDown(self):
        if self._working_directory and os.path.isdir(self._working_directory):
            shutil.rmtree(self._working_directory)

    def working_directory(self):
        self._working_directory = tempfile.mkdtemp()
        return self._working_directory

    def test_general_checks(self):

        m = FindLongestAnimalName()
        m.log_to_stdout = False

        output_file = "{}/animals_summary.json".format(self.working_directory())
        m.animals_output.update(engine_url=f"json://{output_file};indent=4")

        partition_option = m.partition_plea()
        self.assertEqual(2, partition_option.optimal)

        for split_size in [1, 2]:
            slices = m.partition_slice(split_size)
            self.assertEqual(split_size, len(slices))

            all_engine_urls = []
            for s in slices:
                self.assertEqual("find_longest_name", s[0])
                all_engine_urls.extend(s[1]["engine_set"])

            self.assertEqual(2, len(all_engine_urls))
            squashed_urls = " ".join(all_engine_urls)
            self.assertIn("deadly_creatures.csv", squashed_urls)
            self.assertIn("monkeys.tsv", squashed_urls)

    def test_happy_partitioned_path(self):

        m = FindLongestAnimalName()
        m.log_to_stdout = False

        output_file = "{}/animals_summary.json".format(self.working_directory())
        m.animals_output.update(engine_url=f"json://{output_file};indent=4")
        output_encoding = m.animals_output.encoding

        m.go()

        with open(output_file, "r", encoding=output_encoding) as f:
            output_data = json.load(f)

        expected_data = "Crown of thorns starfish"
        self.assertEqual(expected_data, output_data)

    def test_partitioned_connector_resolver(self):
        """
        Key -> Value pairs set on the global ConnectorResolver should be available to worker
        processes.
        Callables aren't yet supported.
        """
        build_context = {"hello_partitioned_context": "important_build_data.ndjson"}
        with ayeaye.connector_resolver.context(**build_context):
            m = DistributedFakeWork()
            m.log_to_stdout = False
            m.go()

        expected_results = set(["important_build_data.ndjson" + str(x) for x in range(10)])
        self.assertEqual(expected_results, set(m.resultset))



from io import StringIO
import json
import os
import shutil
import tempfile
import unittest

import ayeaye

PROJECT_TEST_PATH = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_CSV_PATH = os.path.join(PROJECT_TEST_PATH, "data", "deadly_creatures.csv")


class FakeModel(ayeaye.Model):
    animals = ayeaye.Connect(engine_url=f"csv://{EXAMPLE_CSV_PATH}")

    def build(self):
        for a in self.animals:
            self.log(a.common_name)


class TestModels(unittest.TestCase):
    def setUp(self):
        self._working_directory = None

    def tearDown(self):
        if self._working_directory and os.path.isdir(self._working_directory):
            shutil.rmtree(self._working_directory)

    def working_directory(self):
        self._working_directory = tempfile.mkdtemp()
        return self._working_directory

    def test_go_closes_dataset_connections(self):
        m = FakeModel()
        m.log_to_stdout = False
        m.go()

        self.assertEqual(None, m.animals._file_handle, "File handle should be closed")

    def test_double_usage(self):
        """
        Use Connect as a callable to make a second iterable. Make the cartesian product to
        demonstrate they are independent.
        """

        class AnimalsModel(ayeaye.Model):
            animals_a = ayeaye.Connect(engine_url="csv://" + EXAMPLE_CSV_PATH)
            animals_b = animals_a.clone()
            animals_output = ayeaye.Connect(access=ayeaye.AccessMode.WRITE)

            def build(self):
                cartesian = []
                for a in self.animals_a:
                    for b in self.animals_b:
                        cartesian.append(f"{a.common_name}_{b.common_name}")
                self.animals_output.data = cartesian

        self.assertNotEqual(id(AnimalsModel.animals_a), id(AnimalsModel.animals_b))

        m = AnimalsModel()

        output_file = "{}/animals_summary.json".format(self.working_directory())
        m.animals_output.update(engine_url=f"json://{output_file};indent=4")
        output_encoding = m.animals_output.encoding

        m.go()

        with open(output_file, "r", encoding=output_encoding) as f:
            output_data = json.load(f)

        expected_data = [
            "Crown of thorns starfish_Crown of thorns starfish",
            "Crown of thorns starfish_Golden dart frog",
        ]
        self.assertEqual(expected_data, output_data)

    def test_record_stats(self):
        """
        Use model.stats to store a counter and a set of strings.
        """
        external_log = StringIO()

        m = FakeModel()
        m.log_to_stdout = False
        m.set_logger(external_log)

        # example of common stats when running an ETL
        m.stats["records_processed"] += 100
        # default dict uses 'int' as the default. This check ensures re-assign is possible
        m.stats["categories_seen"] = set(["new_customers", "existing_customers"])
        m.go()

        external_log.seek(0)
        all_the_logs = external_log.read()

        msg = "stat field names and values should be logged when the model finishes"
        self.assertIn("records_processed", all_the_logs, msg)
        self.assertIn("existing_customers", all_the_logs, msg)
